"""
Helper functions for multihtreading news fetching.
"""

from concurrent.futures import ThreadPoolExecutor
from typing import List, Union
import newspaper
from newspaper.article import Article
from newspaper.source import Source


def fetch_news(
    news_list: List[Union[str, Article, Source]], threads: int = 5
) -> List[Union[Article, Source]]:
    """
    Fetch news from a list of sources, articles, or both. Threads will be
    allocated to download and parse the sources or articles. If urls are
    passed into the list, then a new `Article` object will be created for
    it and downloaded + parsed.
    There will be no nlp done on the articles.
    If there is a problem in detecting the language of the urls, then instantiate
    the `Article` object yourself with the language parameter and pass it in.

    Arguments:
        news_list {List[Union[str, Article, Source]]} -- List of sources,
        articles, urls or a mix of them.

        threads {int} -- Number of threads to use for fetching. This affects
        how many items from the news_list are fetched at once. In order to control
        how many threads are used in a `Source` object, use the
        `Configuration`.`number_threads` setting. This could result in
        a high number of threads. Maximum number of threads would be
        `threads` * `Configuration`.`number_threads`.

    """

    def get_item(item: Union[str, Article, Source]) -> Union[Article, Source]:
        if isinstance(item, Article):
            item.download()
            item.parse()
        elif isinstance(item, Source):
            item.download_articles()
            item.parse_articles()
        elif isinstance(item, str):
            item = newspaper.article(url=item)
        else:
            raise TypeError(f"Invalid type {type(item)} for item {item}")

        return item

    with ThreadPoolExecutor(max_workers=threads) as tpe:
        results = tpe.map(get_item, news_list)

    return list(results)



import copy
import re
from statistics import mean
import lxml
import newspaper.extractors.defines as defines
import newspaper.parsers as parsers

score_weights = {
    "start_boosting_score": 1.0,
    "bottom_negativescore_nodes": 0.25,
    "boost_score": 50,
    "parent_node": 1.0,
    "parent_parent_node": 0.4,
    "node_count_threshold": 15,
    "negative_score_threshold": 40,
    "negative_score_boost": 5.0,
}


class ArticleBodyExtractor:
    def __init__(self, config):
        self.config = config
        self.top_node = None
        self.top_node_complemented = None
        self.stopwords_class = config.stopwords_class
        self.language = config.language

    def parse(self, doc: lxml.html.Element):
        """_summary_

        Args:
            doc (lxml.html.Element): _description_
        """
        self.top_node = self.calculate_best_node(doc)
        self.top_node_complemented = self.complement_with_siblings(self.top_node)

    def calculate_best_node(self, doc):
        top_node = None
        nodes_to_check = self.nodes_to_check(doc)
        self.boost_highly_likely_nodes(doc)
        starting_boost = score_weights["start_boosting_score"]

        parent_nodes = []
        nodes_with_text = []

        for node in nodes_to_check:
            text_node = parsers.get_text(node)
            if not text_node:
                continue
            word_stats = self.stopwords_class(
                language=self.language
            ).get_stopword_count(text_node)
            high_link_density = parsers.is_highlink_density(node)
            if word_stats.stop_word_count > 2 and not high_link_density:
                nodes_with_text.append(node)

        nodes_number = len(nodes_with_text)
        negative_scoring = 0
        bottom_negativescore_nodes = (
            float(nodes_number) * score_weights["bottom_negativescore_nodes"]
        )

        for i, node in enumerate(nodes_with_text):
            boost_score = float(0)
            # boost
            if self.is_boostable(node):
                if i >= 0:
                    boost_score = float(
                        (1.0 / starting_boost) * score_weights["boost_score"]
                    )
                    starting_boost += 1
            # nodes_number
            if nodes_number > score_weights["node_count_threshold"]:
                # higher number of possible top nodes
                if (nodes_number - i) <= bottom_negativescore_nodes:
                    booster = float(bottom_negativescore_nodes - (nodes_number - i))
                    boost_score = float(-pow(booster, float(2)))
                    negscore = abs(boost_score) + negative_scoring
                    if negscore > score_weights["negative_score_threshold"]:
                        boost_score = score_weights["negative_score_boost"]

            text_node = parsers.get_text(node)
            word_stats = self.stopwords_class(
                language=self.language
            ).get_stopword_count(text_node)
            upscore = int(word_stats.stop_word_count + boost_score)

            parent_node = node.getparent()
            self.update_score(parent_node, upscore)
            self.update_node_count(parent_node, 1)

            if parent_node not in parent_nodes:
                parent_nodes.append(parent_node)

            # Parent of parent node
            parent_parent_node = parent_node.getparent()
            if parent_parent_node is not None:
                self.update_node_count(parent_parent_node, 1)
                self.update_score(
                    parent_parent_node, upscore * score_weights["parent_parent_node"]
                )
                if parent_parent_node not in parent_nodes:
                    parent_nodes.append(parent_parent_node)

        if parent_nodes:
            parent_nodes.sort(key=parsers.get_node_gravity_score, reverse=True)
            top_node = parent_nodes[0]

        return top_node

    def nodes_to_check(self, doc):
        """Returns a list of nodes we want to search
        on like paragraphs and tables
        """
        nodes_to_check = []
        for tag in ["p", "pre", "td", "article", "div"]:
            if tag == "div":
                items = []
                for attr in [
                    "article-body",
                    "articlebody",
                    "article",
                    "story",
                    "article-content",
                ]:
                    items += parsers.get_tags(
                        doc, tag=tag, attribs={"id": attr}, attribs_match="word"
                    )
                    items += parsers.get_tags(
                        doc, tag=tag, attribs={"class": attr}, attribs_match="word"
                    )
                for class_ in ["paragraph"]:
                    items += parsers.get_tags_regex(
                        doc, tag=tag, attribs={"class": class_}
                    )
                if len(items) == 0 and len(nodes_to_check) < 5:
                    items = parsers.get_tags(doc, tag=tag)
                items = set(items)  # remove duplicates
            else:
                items = parsers.get_tags(doc, tag=tag)
            nodes_to_check += items

        # Do not miss some Article Bodies or Article Sections
        # for itemprop in ['articleBody', 'articlebody', 'articleText',
        #                       'articleSection']:
        #     items = [item  for item in parsers.get_tags(
        #         doc, attribs={"itemprop": itemprop}, attribs_match="word"
        #     ) if item not in nodes_to_check]
        #     nodes_to_check.extend(items)

        return nodes_to_check

    def is_boostable(self, node):
        """A lot of times the first paragraph might be the caption under an image
        so we'll want to make sure if we're going to boost a parent node that
        it should be connected to other paragraphs, at least for the first n
        paragraphs so we'll want to make sure that the next sibling is a
        paragraph and has at least some substantial weight to it.
        """
        para = "p"
        steps_away = 0
        minimum_stopword_count = 5
        max_stepsaway_from_node = 3

        nodes = self.walk_siblings(node)
        for current_node in nodes:
            # <p>
            current_node_tag = current_node.tag
            if current_node_tag == para:
                if steps_away >= max_stepsaway_from_node:
                    return False
                paragraph_text = parsers.get_text(current_node)
                word_stats = self.stopwords_class(
                    language=self.language
                ).get_stopword_count(paragraph_text)
                if word_stats.stop_word_count > minimum_stopword_count:
                    return True
                steps_away += 1
        return False

    def boost_highly_likely_nodes(self, doc: lxml.html.Element):
        """Set a bias score for all nodes under most likely
        article containers. This way we can find articles
        that have little text.

        Args:
            doc (lxml.html.Element): Document to be checked
        """
        candidates = []
        for tag in ["p", "pre", "td", "article", "div"]:
            candidates.extend(parsers.get_tags(doc, tag=tag))

        for e in candidates:
            boost = self.is_highly_likely(e)
            if boost > 0:
                self.update_score(e, boost * score_weights["parent_node"])
                # for child in e.iterdescendants():
                # TODO: find an optimum value
                #     self.update_score(child, boost // 10)

    def is_highly_likely(self, node: lxml.html.Element) -> int:
        """Checks if the node is a well known tag + attributes combination
        for article body containers. This way we can deliver even small
        article bodies with high link density
        see: https://finance.yahoo.com/m/
            0edc9aaa-e3ba-3178-be5a-f8c16fbffff2/warren-buffett-stocks-shaky.html

        Args:
            node (lxml.html.Element): Node to check

        Returns:
            bool: True if node could be an article body top node
        """

        def is_tag_match(node, tag_dict):
            if node.tag != tag_dict.get("tag", node.tag):
                return False
            for k, v in tag_dict.items():
                if k in ["tag", "score_boost"]:
                    continue
                if v.startswith("re:"):
                    v = v[3:]
                    if not re.search(v, node.get(k, ""), re.IGNORECASE):
                        return False
                elif node.get(k, "").lower() != v.lower():
                    return False
            return True

        scores = []
        for tag in defines.ARTICLE_BODY_TAGS:
            if is_tag_match(node, tag):
                scores.append(tag["score_boost"])
        if scores:
            return max(scores)

        return 0

    def update_score(self, node, add_to_score):
        """Adds a score to the gravityScore Attribute we put on divs
        we'll get the current score then add the score we're passing
        in to the current.
        """
        new_score = parsers.get_node_gravity_score(node) + add_to_score
        parsers.set_attribute(node, "gravityScore", str(new_score))

    def update_node_count(self, node, add_to_count):
        """Stores how many decent nodes are under a parent node"""
        new_count = float(node.get("gravityNodes", 0)) + add_to_count
        parsers.set_attribute(node, "gravityNodes", str(new_count))

    def add_siblings(self, top_node):
        res_node = copy.deepcopy(top_node)
        baseline_score = self.get_normalized_score(top_node)
        results = self.walk_siblings(top_node)
        for current_node in results:
            ps = self.get_plausible_content(current_node, baseline_score)
            for p in ps:
                res_node.insert(0, p)
        return res_node

    def get_plausible_content(self, node, baseline_score, score_weight=0.3):
        """Create list of (off-the-tree) paragraphs from a node (most likely a
        sibling of the top node) that are plausible to be part of the
        article body. We use the stop word count as a score of plausibility.
        We accept only paragraphs with a score higher than the baseline score
        weighted by the score_weight parameter.
        The baseline score is a normalized score of the top node.
        """
        if isinstance(
            node, (lxml.etree.CommentBase, lxml.etree.EntityBase, lxml.etree.PIBase)
        ):
            return []

        if node.tag == "p" and node.text and not parsers.is_highlink_density(node):
            element = copy.deepcopy(node)
            element.tail = ""
            return [element]

        paragraphs = parsers.get_tags(node, tag="p")
        result = []
        if not paragraphs:
            return result

        for paragraph in paragraphs:
            text = parsers.get_text(paragraph)
            if not text:
                continue
            if parsers.is_highlink_density(paragraph):
                continue

            word_stats = self.stopwords_class(
                language=self.language
            ).get_stopword_count(text)

            if word_stats.stop_word_count > baseline_score * score_weight:
                element = parsers.create_element(tag="p", text=text)
                result.append(element)

        return result

    def get_normalized_score(self, top_node):
        """We could have long articles that have tons of paragraphs
        so if we tried to calculate the base score against
        the total text score of those paragraphs it would be unfair.
        So we need to normalize the score based on the average scoring
        of the paragraphs within the top node.
        For example if our total score of 10 paragraphs was 1000
        but each had an average value of 100 then 100 should be our base.
        """

        nodes_to_check = parsers.get_tags(top_node, tag="p")

        scores = [parsers.get_node_gravity_score(node) for node in nodes_to_check]
        scores = [score for score in scores if score > 0]  # filter out 0 scores

        return mean(scores) if scores else float("inf")

    def walk_siblings(self, node):
        """returns preceding siblings in reverse order (nearest sibling is first)"""
        return [n for n in node.itersiblings(preceding=True)]

    def complement_with_siblings(self, node: lxml.html.Element) -> lxml.html.Element:
        """Adds surrounding relevant siblings to the top node.
        Attention, it generates off-the-tree node.

        Args:
            node (lxml.html.Element): Top node detected

        Returns:
            lxml.html.Element: off the tree node complemented with siblings
        """
        if node is None:
            return node
        # TODO: test if there is a problem with siblings AFTER the top node
        # node_complemented = self.add_siblings(node)

        node_complemented = self.add_same_level_candidates(node)

        # for e in node_complemented.getchildren():
        #     if e.tag != "p":
        #         if parsers.is_highlink_density(e):
        #             parsers.remove(e)
        return node_complemented

    def add_same_level_candidates(self, node):
        """Adds any siblings that may have a decent score to this node"""
        tree = node.getroottree()

        node_level = parsers.get_level(node)
        # base_score = self.get_normalized_score(node)
        base_score = parsers.get_node_gravity_score(node)

        candidates = parsers.get_nodes_at_level(tree.getroot(), node_level)

        assert node in candidates, "node not in candidates"

        # Create the new tree as a HTML document. Otherwise it will
        # be created as Element tree and cleaners won't work
        new_tree = parsers.fromstring("<html><body></body></html>")
        new_node = new_tree.find("body")

        for n in candidates:
            if n == node:
                new_node.append(copy.deepcopy(node))
                continue

            # avoid adding nodes that do not resemble the top node
            if n.tag != node.tag:
                continue

            score = parsers.get_node_gravity_score(n)

            if score > base_score * 0.3 and not parsers.is_highlink_density(n):
                new_node.append(copy.deepcopy(n))
                continue

            # content_items = self.get_plausible_content(n, base_score)
            # new_node.extend(content_items)

        return new_node



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)
"""
Helper functions for http requests and remote data fetching.
"""
from concurrent.futures import ThreadPoolExecutor
import logging
from typing import Callable, List, Tuple
import requests
from requests import RequestException
from requests import Response
import tldextract

from newspaper import parsers
from newspaper.exceptions import ArticleException, ArticleBinaryDataException
from newspaper.configuration import Configuration

log = logging.getLogger(__name__)

FAIL_ENCODING = "ISO-8859-1"


def do_cache(func: Callable):
    def wrapper(*args, **kwargs):
        if not hasattr(func, "cache"):
            func.cache = {}
        if kwargs.get("url"):
            url = kwargs["url"]
        else:
            url = args[0] if len(args) > 0 else None
        if url:
            tld = tldextract.extract(url).domain + "." + tldextract.extract(url).suffix
        else:
            return func(*args, **kwargs)

        if tld not in func.cache:
            func.cache[tld] = func(*args, **kwargs)
        return func.cache[tld]

    return wrapper


@do_cache
def has_get_ranges(url: str) -> bool:
    """Does this url support HTTP Range requests?"""
    try:
        resp = requests.head(url, timeout=3, allow_redirects=False)
        if resp.status_code in [301, 302, 303, 307, 308]:
            new_url = resp.headers.get("Location")
            if new_url:
                resp = requests.head(url, timeout=3, allow_redirects=True)
                url = new_url

        if "Accept-Ranges" in resp.headers:
            return True

        resp = requests.get(url, headers={"Range": "bytes=0-4"}, timeout=3)
        if resp.status_code == 206:
            return True
    except RequestException as e:
        log.debug("has_get_ranges() error. %s on URL: %s", e, url)
    return False


def is_binary_url(url: str) -> bool:
    """Does this url point to a binary file?"""
    try:
        resp = requests.head(url, timeout=3)
        if "Content-Type" in resp.headers:
            if resp.headers["Content-Type"].startswith("application"):
                if (
                    "json" not in resp.headers["Content-Type"]
                    and "xml" not in resp.headers["Content-Type"]
                ):
                    return True
            if resp.headers["Content-Type"].startswith("image"):
                return True
            if resp.headers["Content-Type"].startswith("video"):
                return True
            if resp.headers["Content-Type"].startswith("audio"):
                return True
            if resp.headers["Content-Type"].startswith("font"):
                return True

        if "Content-Disposition" in resp.headers:
            return True

        if not has_get_ranges(url):
            return False
        resp = requests.get(
            url, headers={"Range": "bytes=0-1000"}, timeout=3, allow_redirects=False
        )
        if resp.status_code in [301, 302, 303, 307, 308]:
            new_url = resp.headers.get("Location")
            if new_url:
                resp = requests.get(
                    new_url,
                    headers={"Range": "bytes=0-1000"},
                    timeout=3,
                    allow_redirects=True,
                )

        content = resp.content
        if isinstance(content, bytes):
            try:
                content = content.decode("utf-8", errors="replace")
            except UnicodeDecodeError:
                pass

        content = content[:1000]

        if len(content) == 0:
            return False

        if "<html" in content:
            return False

        chars = len(
            [
                char
                for char in content
                if 31 < ord(char) < 128 or ord(char) in [9, 10, 13]
            ]
        )
        if chars / len(content) < 0.6:  # 40% of the content is binary
            return True

        return False

    except RequestException as e:
        log.debug("is_binary_url() error. %s on URL: %s", e, url)
    return False


def do_request(url, config):
    if not config.allow_binary_content and has_get_ranges(url):
        if is_binary_url(url):
            raise ArticleBinaryDataException(f"Article is binary data: {url}")

    response = requests.get(
        url=url,
        **config.requests_params,
    )

    return response


def get_html(url, config=None, response=None):
    """Returns the html content from an url.
    if response is provided, no download will occur. The html will be extracted
    from the provided response.
    It does encoding sanitization and dammit if necessary.
    In case of http error (e.g. 404, 500), it returns an empty string.
    If `config`.`http_success_only` is True, it raises an exception in
    case of a http error.
    """
    html = ""
    try:
        html, status_code, _ = get_html_status(url, config, response)
        if status_code >= 400:
            log.warning("get_html() bad status code %s on URL: %s", status_code, url)
            if config.http_success_only:
                raise ArticleException(
                    f"Http error when downloading {url}. Status code: {{status_code}}"
                )
            return ""
    except RequestException as e:
        log.debug("get_html() error. %s on URL: %s", e, url)

    return html


def get_html_status(url, config=None, response=None) -> Tuple[str, int, List[Response]]:
    """Consolidated logic for http requests from newspaper. We handle error cases:
    - Attempt to find encoding of the html by using HTTP header. Fallback to
      'ISO-8859-1' if not provided.
    - Error out if a non 2XX HTTP response code is returned.
    """
    config = config or Configuration()

    if response is not None:
        return (
            _get_html_from_response(response, config),
            response.status_code,
            response.history,
        )

    response = do_request(url, config)

    if response.status_code != 200:
        log.warning(
            "get_html_status(): bad status code %s on URL: %s, html: %s",
            response.status_code,
            url,
            response.text[:200],
        )
    html = _get_html_from_response(response, config)
    if isinstance(html, bytes):
        html = parsers.get_unicode_html(html)

    return html, response.status_code, response.history


def _get_html_from_response(response, config):
    if response.headers.get("content-type") in config.ignored_content_types_defaults:
        return config.ignored_content_types_defaults[
            response.headers.get("content-type")
        ]
    if response.encoding != FAIL_ENCODING:
        # return response as a unicode string
        html = response.text
    else:
        html = response.content
        if "charset" not in response.headers.get("content-type"):
            encodings = requests.utils.get_encodings_from_content(response.text)
            if len(encodings) > 0:
                response.encoding = encodings[0]
                html = response.text

    return html or ""


def multithread_request(urls, config=None):
    """Request multiple urls via mthreading, order of urls & requests is stable
    returns same requests but with response variables filled.
    """
    config = config or Configuration()

    timeout = config.thread_timeout_seconds
    requests_timeout = config.requests_params.get("timeout", 7)

    if timeout < requests_timeout:
        log.warning(
            "multithread_request(): Thread timeout %s < Requests timeout %s, could"
            " cause threads to be stopped before getting the chance to finish. Consider"
            " increasing the thread timeout.",
            timeout,
            requests_timeout,
        )
    results = []
    with ThreadPoolExecutor(max_workers=config.number_threads) as tpe:
        result_futures = [
            tpe.submit(do_request, url=url, config=config) for url in urls
        ]
        for idx, future in enumerate(result_futures):
            url = urls[idx]
            try:
                results.append(future.result())
            except TimeoutError:
                results.append(None)
                log.error("multithread_request(): Thread timeout for URL: %s", url)
            except RequestException as e:
                results.append(None)
                log.warning(
                    "multithread_request(): Http download error %s on URL: %s", e, url
                )

    return results



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

"""
Source objects abstract online news websites & domains. One Source object
can contain multiple Articles. If you want to pull articles from a single
url use the Article object.
Source provdides basic crawling + parsing logic for a news source homepage.
"""

from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
import logging
import re
from typing import List, Optional
from urllib.parse import urljoin, urlsplit, urlunsplit
import lxml

from tldextract import tldextract

from . import network
from . import urls
from . import utils
from .article import Article
from .configuration import Configuration
import newspaper.parsers as parsers
from .extractors import ContentExtractor
from .settings import NUM_THREADS_PER_SOURCE_WARN_LIMIT

log = logging.getLogger(__name__)


@dataclass
class Category:
    """A category object is a representation of a category of news
    on a news source's homepage. For example, on cnn.com, the category
    "world" would be a category object.

    Attributes:
        url(str): The url of the category's homepage. e.g. https://www.cnn.com/world
        html(str): The html of the category's homepage as downloaded by requests.
        doc(lxml.html.HtmlElement): The parsed lxml root of the category's homepage.
    """

    url: str
    html: Optional[str] = None
    doc: Optional[lxml.html.Element] = None


@dataclass
class Feed:
    """A feed object is a representation of an RSS feed on a news source's
    homepage. For example, on cnn.com, the feed
    "http://rss.cnn.com/rss/edition_world.rss" would represent a feed object.
    Attributes:
        url(str): The url of the feed's homepage. e.g. http://rss.cnn.com/rss/edition_world.rss
        rss(str): The rss of the feed's content (xml) as downloaded by requests.
    """

    url: str
    rss: Optional[str] = None
    # TODO self.dom = None, speed up Feedparser


class Source:
    """Sources are abstractions of online news websites such as huffpost or cnn.
    The object will create a list of article urls that belong to the source and
    the list of categories of news (world, politics, etc.) that the source has.
    These categories are inferred from the source's homepage structure.

    Attributes:
        url(str): The url of the source's homepage. e.g. https://www.cnn.com
        config(:any:`Configuration`): The configuration object for this source.
        domain(str): The domain of the source's homepage. e.g. cnn.com
        scheme(str): The scheme of the source's homepage. e.g. https
        categories(list): A list of :any:`Category` objects that belong to the source.
        feeds(list): A list of :any:`Feed` objects that belong to the source containing
            information about the source's RSS feeds.
        articles(list): A list of :any:`Article` objects that belong to the source.
        brand(str): The domain name root of the source. e.g. cnn
        description(str): The description of the source as found in the
            source's meta tags
        doc(lxml.html.HtmlElement): The parsed lxml root of the source's homepage.
        html(str): The html of the source's homepage as downloaded by requests.
        favicon(str): The url of the source's favicon.
        logo_url(str): The url of the source's logo.
    """

    def __init__(
        self,
        url: str,
        read_more_link: Optional[str] = None,
        config: Optional[Configuration] = None,
        **kwargs
    ):
        """The config object for this source will be passed into all of this
        source's children articles unless specified otherwise or re-set.

        Arguments:
            url(str): The url of the source's homepage. e.g. https://www.cnn.com
            read_more_link (str, optional): A xpath selector for the link to the
                full article. make sure that the selector works for all casese,
                not only for one specific article. If needed, you can use
                several xpath selectors separated by `|`. Defaults to "".
            config(:any:`Configuration`, optional): The configuration object
                for this source. Defaults to None.

        Keyword Args:
            **kwargs: Any Configuration class propriety can be overwritten
                    through init keyword  params.
                    Additionally, you can specify any of the following
                    requests parameters:
                    headers, cookies, auth, timeout, allow_redirects,
                    proxies, verify, cert

        """
        if (url is None) or ("://" not in url) or (url[:4] != "http"):
            raise ValueError("Input url is bad!")

        self.config = config or Configuration()
        self.config.update(**kwargs)

        self.extractor = ContentExtractor(self.config)

        self.url = url
        self.url = urls.prepare_url(url)

        self.domain = urls.get_domain(self.url)
        self.scheme = urls.get_scheme(self.url)

        self.categories: List[Category] = []
        self.feeds: List[Feed] = []
        self.articles: List[Article] = []

        self.html = ""
        self.doc = None

        self.logo_url = ""
        self.favicon = ""
        self.brand = tldextract.extract(self.url).domain
        self.description = ""
        self.read_more_link = read_more_link

        self.is_parsed = False
        self.is_downloaded = False

    def build(self):
        """Encapsulates download and basic parsing with lxml. May be a
        good idea to split this into download() and parse() methods.
        """
        self.download()
        self.parse()

        self.set_categories()
        self.download_categories()  # mthread
        self.parse_categories()

        self.set_feeds()
        self.download_feeds()  # mthread
        # self.parse_feeds()

        self.generate_articles()

    def purge_articles(self, reason: str, articles: List[Article]) -> List[Article]:
        """Delete rejected articles, if there is an articles param,
        purge from there, otherwise purge from source instance.

        Reference this StackOverflow post for some of the wonky
        syntax below:
        http://stackoverflow.com/questions/1207406/remove-items-from-a-list-while-iterating-in-python
        """
        if reason == "url":
            articles = [a for a in articles if a.is_valid_url()]
        elif reason == "body":
            articles = [a for a in articles if a.is_valid_body()]
        return articles

    @utils.cache_disk(seconds=86400)
    def _get_category_urls(self, domain):
        """The domain param is **necessary**, since disk caching usese this
        parameter to save the cached categories. Even if it seems unused
        in this method, removing it would render disk_cache useless.
        By default we are caching categories for 1 day.

        You can enable/disable disk_cache in run-time by setting
            utils.cache_disk.enabled = True/False
        """
        return self.extractor.get_category_urls(self.url, self.doc)

    def set_categories(self):
        utils.cache_disk.enabled = not self.config.disable_category_cache
        urls = self._get_category_urls(self.domain)
        self.categories = [Category(url=url) for url in set(urls)]

    def set_feeds(self):
        """Don't need to cache getting feed urls, it's almost
        instant with xpath
        """
        common_feed_urls = ["/feed", "/feeds", "/rss"]
        common_feed_urls = [urljoin(self.url, url) for url in common_feed_urls]

        split = urlsplit(self.url)
        if split.netloc in ("medium.com", "www.medium.com"):
            # should handle URL to user or user's post
            if split.path.startswith("/@"):
                new_path = "/feed/" + split.path.split("/")[1]
                new_parts = split.scheme, split.netloc, new_path, "", ""
                common_feed_urls.append(urlunsplit(new_parts))

        common_feed_urls_as_categories = [Category(url=url) for url in common_feed_urls]

        category_urls = [c.url for c in common_feed_urls_as_categories]
        responses = network.multithread_request(category_urls, self.config)

        for response, feed in zip(responses, common_feed_urls_as_categories):
            if response and response.status_code < 400:
                feed.html = network.get_html(feed.url, response=response)

        # Remove empty or erroneous feeds
        common_feed_urls_as_categories = [
            c for c in common_feed_urls_as_categories if c.html
        ]

        for _ in common_feed_urls_as_categories:
            doc = parsers.fromstring(_.html)
            _.doc = doc

        common_feed_urls_as_categories = [
            c for c in common_feed_urls_as_categories if c.doc is not None
        ]

        categories_and_common_feed_urls = (
            self.categories + common_feed_urls_as_categories
        )
        urls = self.extractor.get_feed_urls(self.url, categories_and_common_feed_urls)
        self.feeds = [Feed(url=url) for url in urls]

    def set_description(self):
        """Sets a blurb for this source, for now we just query the
        desc html attribute
        """
        metadata = self.extractor.get_metadata(self.url, self.doc)
        self.description = metadata["description"]

    def download(self):
        """Downloads html of source"""
        self.html = network.get_html(self.url, self.config)

    def download_categories(self):
        """Download all category html, can use mthreading"""
        category_urls = self.category_urls()
        responses = network.multithread_request(category_urls, self.config)

        for response, category in zip(responses, self.categories):
            if response and response.status_code < 400:
                category.html = network.get_html(category.url, response=response)

        self.categories = [c for c in self.categories if c.html]

        return self.categories

    def download_feeds(self):
        """Download all feed html, can use mthreading"""
        feed_urls = self.feed_urls()
        responses = network.multithread_request(feed_urls, self.config)

        for response, feed in zip(responses, self.feeds):
            if response and response.status_code < 400:
                feed.rss = network.get_html(feed.url, response=response)
        self.feeds = [f for f in self.feeds if f.rss]
        return self.feeds

    def parse(self):
        """Sets the lxml root, also sets lxml roots of all
        children links, also sets description
        """
        # TODO: This is a terrible idea, ill try to fix it when i'm more rested
        self.doc = parsers.fromstring(self.html)
        if self.doc is None:
            log.warning("Source %s parse error.", self.url)
            return
        self.set_description()

    def parse_categories(self):
        """Parse out the lxml root in each category"""
        log.debug("We are extracting from %d categories", self.categories)
        for category in self.categories:
            doc = parsers.fromstring(category.html)
            category.doc = doc

        self.categories = [c for c in self.categories if c.doc is not None]

    def _map_title_to_feed(self, feed):
        doc = parsers.fromstring(feed.rss)
        if doc is None:
            # http://stackoverflow.com/a/24893800
            return None

        elements = parsers.get_tags(doc, tag="title")
        feed.title = next(
            (element.text for element in elements if element.text), self.brand
        )
        return feed

    def parse_feeds(self):
        """Add titles to feeds"""
        log.debug("We are parsing %d feeds", self.feeds)
        self.feeds = [self._map_title_to_feed(f) for f in self.feeds]

    def feeds_to_articles(self) -> List[Article]:
        """Returns a list of :any:`Article` objects based on
        articles found in the Source's RSS feeds"""
        articles = []

        def get_urls(feed):
            feed = re.sub("<[^<]+?>", " ", str(feed))
            results = re.findall(
                r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|"
                "(?:%[0-9a-fA-F][0-9a-fA-F]))+",
                feed,
            )
            results = [x.strip() for x in results]
            return results

        for feed in self.feeds:
            urls = get_urls(feed.rss)
            cur_articles = []
            before_purge = len(urls)

            for url in urls:
                article = Article(
                    url=url,
                    source_url=feed.url,
                    read_more_link=self.read_more_link,
                    config=self.config,
                )
                cur_articles.append(article)

            cur_articles = self.purge_articles("url", cur_articles)
            after_purge = len(cur_articles)

            if self.config.memorize_articles:
                cur_articles = utils.memorize_articles(self, cur_articles)
            after_memo = len(cur_articles)

            articles.extend(cur_articles)

            log.debug(
                "%d->%d->%d for %s", before_purge, after_purge, after_memo, feed.url
            )
        return articles

    def categories_to_articles(self) -> List[Article]:
        """Takes the categories, splays them into a big list of urls and churns
        the articles out of each url with the url_to_article method
        """
        articles = []

        def get_urls(doc):
            if doc is None:
                return []
            return [
                (a.get("href"), a.text)
                for a in parsers.get_tags(doc, tag="a")
                if a.get("href")
            ]

        for category in self.categories:
            cur_articles = []
            url_title_tups = get_urls(category.doc)
            before_purge = len(url_title_tups)

            for tup in url_title_tups:
                indiv_url = tup[0]
                indiv_title = tup[1]

                _article = Article(
                    url=indiv_url,
                    source_url=category.url,
                    read_more_link=self.read_more_link,
                    title=indiv_title,
                    config=self.config,
                )
                cur_articles.append(_article)

            cur_articles = self.purge_articles("url", cur_articles)
            after_purge = len(cur_articles)

            if self.config.memorize_articles:
                cur_articles = utils.memorize_articles(self, cur_articles)
            after_memo = len(cur_articles)

            articles.extend(cur_articles)

            log.debug(
                "%d->%d->%d for %s", before_purge, after_purge, after_memo, category.url
            )
        return articles

    def _generate_articles(self):
        """Returns a list of all articles, from both categories and feeds"""
        category_articles = self.categories_to_articles()
        feed_articles = self.feeds_to_articles()

        articles = feed_articles + category_articles
        uniq = {article.url: article for article in articles}
        return list(uniq.values())

    def generate_articles(self, limit=5000):
        """Saves all current articles of news source, filter out bad urls"""
        articles = self._generate_articles()
        self.articles = articles[:limit]
        log.debug("%d articles generated and cutoff at %d", len(articles), limit)

    def download_articles(self) -> List[Article]:
        """Starts the ``download()`` for all :any:`Article` objects
        from the ``articles`` property. It can run single threaded or
        multi-threaded.
        Returns:
            List[:any:`Article`]: A list of downloaded articles.
        """
        urls = self.article_urls()
        failed_articles = []

        threads = self.config.number_threads

        if threads > NUM_THREADS_PER_SOURCE_WARN_LIMIT:
            log.warning(
                "Using %s+ threads on a single source may result in rate limiting!",
                NUM_THREADS_PER_SOURCE_WARN_LIMIT,
            )
        responses = network.multithread_request(urls, self.config)
        # Note that the responses are returned in original order
        with ThreadPoolExecutor(max_workers=threads) as tpe:
            futures = []
            for response, article in zip(responses, self.articles):
                if response and response.status_code < 400:
                    html = network.get_html(article.url, response=response)
                else:
                    html = ""
                    failed_articles.append(article.url)

                futures.append(tpe.submit(article.download, input_html=html))

            self.articles = [future.result() for future in futures]

        self.is_downloaded = True

        if len(failed_articles) > 0:
            log.warning(
                "There were %d articles that failed to download: %s",
                len(failed_articles),
                ", ".join(failed_articles),
            )
        return self.articles

    def parse_articles(self):
        """Parse all articles, delete if too small"""
        for article in self.articles:
            article.parse()

        self.articles = self.purge_articles("body", self.articles)
        self.is_parsed = True

    def size(self):
        """Returns the number of articles linked to this news source"""
        if self.articles is None:
            return 0
        return len(self.articles)

    def clean_memo_cache(self):
        """Clears the memoization cache for this specific news domain"""
        utils.clear_memo_cache(self)

    def feed_urls(self):
        """Returns a list of feed urls"""
        return [feed.url for feed in self.feeds]

    def category_urls(self):
        """Returns a list of category urls"""
        return [category.url for category in self.categories]

    def article_urls(self):
        """Returns a list of article urls"""
        return [article.url for article in self.articles]

    def print_summary(self):
        """Prints out a summary of the data in our source instance"""
        print("[source url]:", self.url)
        print("[source brand]:", self.brand)
        print("[source domain]:", self.domain)
        print("[source len(articles)]:", len(self.articles))
        print("[source description[:50]]:", self.description[:50])

        print("printing out 10 sample articles...")

        for a in self.articles[:10]:
            print("\t", "[url]:", a.url)
            print("\t[title]:", a.title)
            print("\t[len of text]:", len(a.text))
            print("\t[keywords]:", a.keywords)
            print("\t[len of html]:", len(a.html))
            print("\t==============")

        print("feed_urls:", self.feed_urls())
        print("\r\n")
        print("category_urls:", self.category_urls())



import re
from typing import Any, Dict, Optional, Set, Union
from urllib.parse import urlparse, urlunparse

import lxml
from newspaper.configuration import Configuration
import newspaper.parsers as parsers
from newspaper.extractors.defines import (
    A_HREF_TAG_SELECTOR,
    A_REL_TAG_SELECTOR,
    META_LANGUAGE_TAGS,
    RE_LANG,
)


class MetadataExtractor:
    def __init__(self, config: Configuration) -> None:
        self.config = config
        self.meta_data: Dict[str, Any] = {
            "language": None,
            "type": None,
            "canonical_link": None,
            "site_name": None,
            "description": None,
            "keywords": None,
            "tags": None,
            "data": None,
        }

    def parse(self, article_url: str, doc: lxml.html.Element) -> Dict[str, Any]:
        """Parse the article's HTML for any known metadata attributes"""
        self.meta_data["language"] = self._get_meta_language(doc)
        self.meta_data["type"] = self._get_meta_field(doc, "og:type")
        self.meta_data["canonical_link"] = self._get_canonical_link(article_url, doc)
        self.meta_data["site_name"] = self._get_meta_field(doc, "og:site_name")
        self.meta_data["description"] = self._get_meta_field(
            doc, ["description", "og:description"]
        )
        self.meta_data["keywords"] = [
            k.strip() for k in self._get_meta_field(doc, "keywords").split(",")
        ]
        self.meta_data["data"] = self._get_metadata(doc)

        return self.meta_data

    def _get_meta_language(self, doc: lxml.html.Element) -> Optional[str]:
        """Return the language string of the article, or None if it cannot be
        determined.
        """

        def get_if_valid(s: str) -> Optional[str]:
            if s is None or len(s) < 2:
                return None
            s = s[:2]
            if re.search(RE_LANG, s):
                return s.lower()
            return None

        attr = get_if_valid(parsers.get_attribute(doc, "lang"))
        if attr:
            return attr

        for elem in META_LANGUAGE_TAGS:
            meta_tag = parsers.get_tags(
                doc, tag=elem["tag"], attribs={elem["attr"]: elem["value"]}
            )

            if meta_tag:
                attr = get_if_valid(meta_tag[0])
                if attr:
                    return attr

        return None

    def _get_canonical_link(
        self, article_url: str, doc: lxml.html.Element
    ) -> Optional[str]:
        """Return the article's canonical URL

        Gets the first available value of:
        1. The rel=canonical tag
        2. The og:url tag
        """
        candidates = [
            node.get("href")
            for node in parsers.get_tags(doc, tag="link", attribs={"rel": "canonical"})
        ]

        candidates.append(self._get_meta_field(doc, "og:url"))
        candidates = [c.strip() for c in candidates if c and c.strip()]

        if candidates:
            meta_url = candidates[0]
            parsed_meta_url = urlparse(meta_url)
            if not parsed_meta_url.hostname:
                # MIGHT not have a hostname in meta_url
                # parsed_url.path might be 'example.com/article.html' where
                # clearly example.com is the hostname
                parsed_article_url = urlparse(article_url)
                strip_hostname_in_meta_path = re.match(
                    rf".*{parsed_article_url.hostname}(?=/)/(.*)",
                    parsed_meta_url.path,
                )
                if strip_hostname_in_meta_path:
                    true_path = strip_hostname_in_meta_path.group(1)
                else:
                    true_path = parsed_meta_url.path

                # true_path may contain querystrings and fragments
                meta_url = urlunparse(
                    (
                        parsed_article_url.scheme,
                        parsed_article_url.hostname,
                        true_path,
                        "",
                        "",
                        "",
                    )
                )
            return meta_url

        return None

    def _get_metadata(self, doc: lxml.html.Element) -> Dict[str, Any]:
        """Extracts metadata from the article's HTML"""
        data: Dict[str, Any] = {}
        properties = parsers.get_tags(doc, "meta")
        for prop in properties:
            key = prop.attrib.get("property") or prop.attrib.get("name")
            value = prop.attrib.get("content") or prop.attrib.get("value")

            if not key or not value:
                continue

            key, value = key.strip(), value.strip()
            if value.isdigit():
                value = int(value)

            if ":" not in key:
                data[key] = value
                continue

            key = key.split(":")
            key_head = key.pop(0)
            ref = data.get(key_head, {})

            if isinstance(ref, str) or isinstance(ref, int):
                data[key_head] = {key_head: ref}
                ref = data[key_head]

            for idx, part in enumerate(key):
                if idx == len(key) - 1:
                    ref[part] = value
                    break
                if not ref.get(part):
                    ref[part] = dict()
                elif isinstance(ref.get(part), (str, int)):
                    # Not clear what to do in this scenario,
                    # it's not always a URL, but an ID of some sort
                    ref[part] = {"identifier": ref[part]}
                ref = ref[part]
        return data

    def _get_tags(self, doc: lxml.html.Element) -> Set[str]:
        """Extracts tags from the article's HTML"""

        elements = doc.xpath(A_HREF_TAG_SELECTOR)
        elements += doc.xpath(A_REL_TAG_SELECTOR)

        if not elements:
            return set()

        tags = [parsers.get_text(el) for el in elements if parsers.get_text(el)]
        return set(tags)

    def _get_meta_field(self, doc: lxml.html.Element, fields: Union[str, list]) -> str:
        """Extract a given meta field from document."""
        if isinstance(fields, str):
            fields = [fields]
        for f in fields:
            meta_fields = parsers.get_metatags(doc, value=f)
            for meta_field in meta_fields:
                val = meta_field.get("content", "").strip()
                if val:
                    return val
        return ""



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

"""
This class holds configuration objects, which can be thought of
as settings.py but dynamic and changing for whatever parent object
holds them. For example, pass in a config object to an Article
object, Source object, or even network methods, and it just works.
"""
import logging

from warnings import warn
from http.cookiejar import CookieJar as cj

from newspaper.utils import get_available_languages

from .text import (
    StopWords,
    StopWordsArabic,
    StopWordsChinese,
    StopWordsKorean,
    StopWordsHindi,
    StopWordsJapanese,
    StopWordsThai,
)
from .version import __version__

log = logging.getLogger(__name__)


class Configuration:
    """Modifies Article / Source properties.
    Attributes:
        min_word_count (int): minimum number of word tokens in an article text
        min_sent_count (int): minimum number of sentences in an article text
        max_title (int): :any:`Article.title` max number of chars. ``title``
            is truncated to this length
        max_text (int): :any:`Article.text` max number of chars. ``text`` is
            truncated to this length
        max_keywords (int): maximum number of keywords inferred
            by :any:`Article.nlp()`
        max_authors (int): maximum number of authors returned
            in :any:`Article.authors`
        max_summary (int): max number of chars in :any:`Article.summary`,
            truncated to this length
        max_summary_sent (int): maximum number of sentences
            in :any:`Article.summary`
        max_file_memo (int): max number of urls we cache for each news source
        top_image_settings (dict): settings for finding top
            image. You can set the following:
                * ``min_width``: minimum width of image (default 300) in
                    order to be considered top image
                * ``min_height``: minimum height of image (default 200) in
                    order to be considered top image
                * ``min_area``: minimum area of image (default 10000) in
                    order to be considered top image
                * ``max_retries``: maximum number of retries to download
                    the image (default 2)
        memorize_articles (bool): If True, it will cache and save
            articles run between runs. The articles are *NOT* cached.
            It will save the parsed article urls between different
            `Source`.`generate_articles()` runs. default True.
        disable_category_cache (bool): If True, it will not cache the `Source`
            category urls. default False.
        fetch_images (bool): If False, it will not download images
            to verify if they obide by the settings in top_image_settings.
            default True.
        follow_meta_refresh (bool): if True, it will follow meta refresh
            redirect when downloading an article. default False.
        clean_article_html (bool): if True it will clean 'unnecessary' tags
            from the article body html.
            Affected property is :any:`Article.article_html`.
            Default True.
        http_success_only (bool): if True, it will raise an ``ArticleException``
             if the html status_code is >= 400 (e.g. 404 page). default True.
        stopwords_class (obj): unique stopword classes for oriental languages,
            don't toggle
        requests_params (dict): Any of the params for the
            `get call`_ from ``requests`` library
        number_threads (int): number of threads to use for multi-threaded downloads
        verbose (bool): if True, it will output debugging information
            **deprecated**: Use the standard python logging module instead
        thread_timeout_seconds (int): timeout for threads
        allow_binary_content (bool): if True, it will allow binary content
            to be downloaded and stored in :any:`Article.html`. Allowing
            this for Source building can lead to longer processing times
            and could hang the process due to huge binary files (such as movies)
            default False.
        ignored_content_types_defaults (dict): dictionary of content-types
            and a default stub content.
            These content type will not be downloaded.
            **Note:**
             If `allow_binary_content` is False,
            binary content will lead to `ArticleBinaryDataException` for
            `Article.download()` and will be skipped in `Source.build()`. This
            will override the defaults in :any:`ignored_content_types_defaults`
            if these match binary files.
        use_cached_categories (bool): if set to False, the cached categories
            will be ignored and a the :any:`Source` will recompute the category
             list every time you build it.
        MIN_WORD_COUNT (int):
            .. deprecated:: 0.9.2
                use :any:`Configuration.min_word_count` instead
        MIN_SENT_COUNT (int):
            .. deprecated:: 0.9.2
                use :any:`Configuration.min_sent_count` instead
        MAX_TITLE (int):
            .. deprecated:: 0.9.2
                use :any:`Configuration.max_title` instead
        MAX_TEXT (int):
            .. deprecated:: 0.9.2
                use :any:`Configuration.max_text` instead
        MAX_KEYWORDS (int):
            .. deprecated:: 0.9.2
                use :any:`Configuration.max_keywords` instead
        MAX_AUTHORS (int):
            .. deprecated:: 0.9.2
                use :any:`Configuration.max_authors` instead
        MAX_SUMMARY (int):
            .. deprecated:: 0.9.2
                use :any:`Configuration.max_summary` instead
        MAX_SUMMARY_SENT (int):
            .. deprecated:: 0.9.2
                use :any:`Configuration.max_summary_sent` instead
        MAX_FILE_MEMO (int):
            .. deprecated:: 0.9.2
                use :any:`Configuration.max_file_memo` instead

    .. _get call:
        https://requests.readthedocs.io/en/latest/api/#requests.get
    """

    def __init__(self):
        """
        Modify any of these Article / Source properties
        TODO: Have a separate ArticleConfig and SourceConfig extend this!
        """
        self.min_word_count = 300  # num of word tokens in text
        self.min_sent_count = 7  # num of sentence tokens
        self.max_title = 200  # num of chars
        self.max_text = 100000  # num of chars
        self.max_keywords = 35  # num of strings in list
        self.max_authors = 10  # num strings in list
        self.max_summary = 5000  # num of chars
        self.max_summary_sent = 5  # num of sentences

        # max number of urls we cache for each news source
        self.max_file_memo = 20000

        self.top_image_settings = {
            "min_width": 300,
            "min_height": 200,
            "min_area": 10000,
            "max_retries": 2,
        }

        # Cache and save articles run after run
        self.memorize_articles = True

        # If true, it will not cache the `Source` category urls
        self.disable_category_cache = False

        # Set this to false if you don't care about getting images
        self.fetch_images = True

        # Follow meta refresh redirect when downloading
        self.follow_meta_refresh = False

        # Don't toggle this variable, done internally
        self._use_meta_language = True

        # You may keep the html of just the main article body
        self.clean_article_html = True

        # Fail for error responses (e.g. 404 page)
        self.http_success_only = True

        # English is the fallback
        self._language = "en"

        # Unique stopword classes for oriental languages, don't toggle
        self.stopwords_class = StopWords

        # Params for get call from `requests` lib
        self.requests_params = {
            "timeout": 7,
            "proxies": {},
            "headers": {
                "User-Agent": f"newspaper/{__version__}",
            },
            "cookies": cj(),
        }

        # Number of threads to use for mthreaded downloads
        self.number_threads = 10

        # Deprecated, use standard python logging module instead (debug level)
        self.verbose = False  # for debugging

        self.thread_timeout_seconds = 10

        self.allow_binary_content = False

        self.ignored_content_types_defaults = {}

    def update(self, **kwargs):
        """Update the configuration object with the given keyword arguments.
        Arguments:
                **kwargs: The keyword arguments to update.
        """

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def browser_user_agent(self):
        """str: The user agent string sent to web servers when downloading
        articles. If not set, it will default to the following: newspaper/x.x.x
        i.e. newspaper/0.9.1"""
        if "headers" not in self.requests_params:
            self.requests_params["headers"] = {}
        return self.requests_params["headers"].get("User-Agent")

    @browser_user_agent.setter
    def browser_user_agent(self, value):
        if "headers" not in self.requests_params:
            self.requests_params["headers"] = {}
        self.requests_params["headers"]["User-Agent"] = value

    @property
    def headers(self):
        """str: The headers sent to web servers when downloading articles.
        It will set the headers for the `get call`_ from ``requests`` library.
        **Note**: If you set the :any:`browser_user_agent` property, it will
        override the ``User-Agent`` header."""
        return self.requests_params.get("headers")

    @headers.setter
    def headers(self, value):
        self.requests_params["headers"] = value

    @property
    def request_timeout(self):
        """Optional[int,Tuple[int,int]]: The timeout for the `get call`_
        from ``requests`` library. If not set, it will default to 7 seconds."""
        return self.requests_params.get("timeout")

    @request_timeout.setter
    def request_timeout(self, value):
        self.requests_params["timeout"] = value

    @property
    def proxies(self):
        """Optional[dict]: The proxies for the `get call`_ from ``requests``
        library. If not set, it will default to no proxies."""
        return self.requests_params.get("proxies")

    @proxies.setter
    def proxies(self, value):
        self.requests_params["proxies"] = value

    @property
    def language(self):
        """str: the iso-639-1 two letter code of the language.
        If not set, :any:`Article` will try to use the meta information of the webite
        to get the language. english is the fallback"""

        return self._language

    @language.setter
    def language(self, value: str):
        if not value or len(value) != 2:
            raise ValueError(
                "Your input language must be a 2 char language code,                "
                " for example: english-->en \n and german-->de"
            )
        if value not in list(get_available_languages()):
            raise ValueError(
                f"We do not currently support input language {value} yet"
                "supported languages are: {get_available_languages()}"
            )

        # If explicitly set language, don't use meta
        self._use_meta_language = False

        # Set oriental language stopword class
        self._language = value
        self.stopwords_class = self.get_stopwords_class(value)

    @property
    def use_meta_language(self):
        """Read-only property that indicates whether the meta language
        read from the website was used or the language was explicitly set.
        Returns:
            bool: True if the meta language was used, False if the language
            was explicitly set.
        """
        return self._use_meta_language

    @staticmethod
    def get_stopwords_class(language: str):
        """Get the stopwords class for the given language.
        Arguments:
            language (str): The language for which it will return the StopWords object.
        Returns:
            class(StopWords): The stopwords class for the given language.
        """
        if language == "ko":
            return StopWordsKorean
        elif language == "hi":
            return StopWordsHindi
        elif language == "zh":
            return StopWordsChinese
        # Persian and Arabic Share an alphabet
        # There is a persian parser https://github.com/sobhe/hazm,
        # but nltk is likely sufficient
        elif language == "ar" or language == "fa":
            return StopWordsArabic
        elif language == "ja":
            return StopWordsJapanese
        elif language == "th":
            return StopWordsThai
        return StopWords

    @property
    def MIN_WORD_COUNT(self):
        warn(
            "`MIN_WORD_COUNT` is deprecated, use `min_word_count` instead",
            DeprecationWarning,
        )
        return self.min_word_count

    @MIN_WORD_COUNT.setter
    def MIN_WORD_COUNT(self, value):
        warn(
            "`MIN_WORD_COUNT` is deprecated, use `min_word_count` instead",
            DeprecationWarning,
        )
        self.min_word_count = value

    @property
    def MIN_SENT_COUNT(self):
        warn(
            "`MIN_SENT_COUNT` is deprecated, use `min_sent_count` instead",
            DeprecationWarning,
        )
        return self.min_sent_count

    @MIN_SENT_COUNT.setter
    def MIN_SENT_COUNT(self, value):
        warn(
            "`MIN_SENT_COUNT` is deprecated, use `min_sent_count` instead",
            DeprecationWarning,
        )
        self.min_sent_count = value

    @property
    def MAX_TITLE(self):
        warn("`MAX_TITLE` is deprecated, use `max_title` instead", DeprecationWarning)
        return self.max_title

    @MAX_TITLE.setter
    def MAX_TITLE(self, value):
        warn("`MAX_TITLE` is deprecated, use `max_title` instead", DeprecationWarning)
        self.max_title = value

    @property
    def MAX_TEXT(self):
        warn("`MAX_TEXT` is deprecated, use `max_text` instead", DeprecationWarning)
        return self.max_text

    @MAX_TEXT.setter
    def MAX_TEXT(self, value):
        warn("`MAX_TEXT` is deprecated, use `max_text` instead", DeprecationWarning)
        self.max_text = value

    @property
    def MAX_KEYWORDS(self):
        warn(
            "`MAX_KEYWORDS` is deprecated, use `max_keywords` instead",
            DeprecationWarning,
        )
        return self.max_keywords

    @MAX_KEYWORDS.setter
    def MAX_KEYWORDS(self, value):
        warn(
            "`MAX_KEYWORDS` is deprecated, use `max_keywords` instead",
            DeprecationWarning,
        )
        self.max_keywords = value

    @property
    def MAX_AUTHORS(self):
        warn(
            "`MAX_AUTHORS` is deprecated, use `max_authors` instead", DeprecationWarning
        )
        return self.max_authors

    @MAX_AUTHORS.setter
    def MAX_AUTHORS(self, value):
        warn(
            "`MAX_AUTHORS` is deprecated, use `max_authors` instead", DeprecationWarning
        )
        self.max_authors = value

    @property
    def MAX_SUMMARY(self):
        warn(
            "`MAX_SUMMARY` is deprecated, use `max_summary` instead", DeprecationWarning
        )
        return self.max_summary

    @MAX_SUMMARY.setter
    def MAX_SUMMARY(self, value):
        warn(
            "`MAX_SUMMARY` is deprecated, use `max_summary` instead", DeprecationWarning
        )
        self.max_summary = value

    @property
    def MAX_SUMMARY_SENT(self):
        warn(
            "`MAX_SUMMARY_SENT` is deprecated, use `max_summary_sent` instead",
            DeprecationWarning,
        )
        return self.max_summary_sent

    @MAX_SUMMARY_SENT.setter
    def MAX_SUMMARY_SENT(self, value):
        warn(
            "`MAX_SUMMARY_SENT` is deprecated, use `max_summary_sent` instead",
            DeprecationWarning,
        )
        self.max_summary_sent = value

    @property
    def MAX_FILE_MEMO(self):
        warn(
            "`MAX_FILE_MEMO` is deprecated, use `max_file_memo` instead",
            DeprecationWarning,
        )
        return self.max_file_memo

    @MAX_FILE_MEMO.setter
    def MAX_FILE_MEMO(self, value):
        warn(
            "`MAX_FILE_MEMO` is deprecated, use `max_file_memo` instead",
            DeprecationWarning,
        )
        self.max_file_memo = value



import re
import lxml

from newspaper.configuration import Configuration
import newspaper.parsers as parsers
from newspaper.extractors.defines import (
    MOTLEY_REPLACEMENT,
    TITLE_META_INFO,
    TITLE_REPLACEMENTS,
)


class TitleExtractor:
    def __init__(self, config: Configuration) -> None:
        self.config = config
        self.title: str = ""

    def parse(self, doc: lxml.html.Element) -> str:
        """Fetch the article title and analyze it

        Assumptions:
        - title tag is the most reliable (inherited from Goose)
        - h1, if properly detected, is the best (visible to users)
        - og:title and h1 can help improve the title extraction
        - python == is too strict, often we need to compare filtered
          versions, i.e. lowercase and ignoring special chars

        Explicit rules:
        1. title == h1, no need to split
        2. h1 similar to og:title, use h1
        3. title contains h1, title contains og:title, len(h1) > len(og:title), use h1
        4. title starts with og:title, use og:title
        5. use title, after splitting
        """
        self.title = ""
        title_element = parsers.get_tags(doc, tag="title")
        # no title found
        if title_element is None or len(title_element) == 0:
            return self.title

        # title elem found
        title_text = parsers.get_text(title_element[0])
        used_delimeter = False

        # title from h1
        # - extract the longest text from all h1 elements
        # - too short texts (fewer than 2 words) are discarded
        # - clean double spaces
        title_text_h1 = ""
        title_element_h1_list = parsers.get_tags(doc, tag="h1") or []
        title_text_h1_list = [parsers.get_text(tag) for tag in title_element_h1_list]
        if title_text_h1_list:
            # sort by len and set the longest
            title_text_h1_list.sort(key=len, reverse=True)
            title_text_h1 = title_text_h1_list[0]
            # discard too short texts
            if len(title_text_h1.split(" ")) <= 2:
                title_text_h1 = ""
            # clean double spaces
            title_text_h1 = " ".join([x for x in title_text_h1.split() if x])

        # title from og:title
        def get_fb_title():
            for known_meta_tag in TITLE_META_INFO:
                meta_tags = parsers.get_metatags(doc, value=known_meta_tag)
                for meta_tag in meta_tags:
                    title_text_fb = meta_tag.get("content", "").strip()
                    if title_text_fb:
                        return title_text_fb
            return ""

        title_text_fb = get_fb_title()

        # create filtered versions of title_text, title_text_h1, title_text_fb
        # for finer comparison
        filter_regex = re.compile(r"[^\u4e00-\u9fa5a-zA-Z0-9\ ]")
        filter_title_text = filter_regex.sub("", title_text).lower()
        filter_title_text_h1 = filter_regex.sub("", title_text_h1).lower()
        filter_title_text_fb = filter_regex.sub("", title_text_fb).lower()

        # check for better alternatives for title_text and possibly skip splitting
        if title_text_h1 == title_text:
            used_delimeter = True
        elif filter_title_text_h1 and filter_title_text_h1 == filter_title_text_fb:
            title_text = title_text_h1
            used_delimeter = True
        elif (
            filter_title_text_h1
            and filter_title_text_h1 in filter_title_text
            and filter_title_text_fb
            and filter_title_text_fb in filter_title_text
            and len(title_text_h1) > len(title_text_fb)
        ):
            title_text = title_text_h1
            used_delimeter = True
        elif (
            filter_title_text_fb
            and filter_title_text_fb != filter_title_text
            and filter_title_text.startswith(filter_title_text_fb)
        ):
            title_text = title_text_fb
            used_delimeter = True

        if not used_delimeter:
            for delimiter in ["|", "-", "_", "/", " » "]:
                if delimiter in title_text:
                    title_text = self._split_title(title_text, delimiter, title_text_h1)
                    used_delimeter = True
                    break

        title = title_text.replace(*MOTLEY_REPLACEMENT)

        # in some cases the final title is quite similar to title_text_h1
        # (either it differs for case, for special chars, or it's truncated)
        # in these cases, we prefer the title_text_h1
        filter_title = filter_regex.sub("", title).lower()
        if filter_title_text_h1 == filter_title:
            title = title_text_h1

        self.title = title.strip()

        return self.title

    def _split_title(self, title: str, delimiter: str, hint: str = None):
        """Split the title to best part possible"""
        large_text_length = 0
        large_text_index = 0
        title_pieces = title.split(delimiter)

        if hint:
            filter_regex = re.compile(r"[^a-zA-Z0-9\ ]")
            hint = filter_regex.sub("", hint).lower()

        # find the largest title piece
        for i, title_piece in enumerate(title_pieces):
            current = title_piece.strip()
            if hint and hint in filter_regex.sub("", current).lower():
                large_text_index = i
                break
            if len(current) > large_text_length:
                large_text_length = len(current)
                large_text_index = i

        # replace content
        title = title_pieces[large_text_index]
        return title.replace(*TITLE_REPLACEMENTS)



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

"""
Holds the code for cleaning out unwanted tags from the lxml
dom xpath.
"""
import re
import newspaper.parsers as parsers


class DocumentCleaner:
    def __init__(self, config):
        """Set appropriate tag names and regexes of tags to remove
        from the HTML
        """
        self.config = config
        self.remove_nodes_re = (
            "^side$|combx|retweet|mediaarticlerelated|menucontainer|"
            "navbar|storytopbar-bucket|utility-bar|inline-share-tools"
            "|comment|PopularQuestions|contact|foot|footer|Footer|footnote"
            "|cnn_strycaptiontxt|cnn_html_slideshow|cnn_strylftcntnt"
            "|links|meta$|shoutbox|sponsor"
            "|tags|socialnetworking|socialNetworking|cnnStryHghLght"
            "|cnn_stryspcvbx|^inset$|pagetools|post-attributes"
            "|welcome_form|contentTools2|the_answers"
            "|communitypromo|runaroundLeft|subscribe|vcard|articleheadings"
            "|date|^print$|popup|author-dropdown|tools|socialtools|byline"
            "|konafilter|KonaFilter|breadcrumbs|^fn$|wp-caption-text"
            "|legende|ajoutVideo|timestamp|js_replies"
        )
        self.remove_nodes_related_re = (
            r"related[-\s\_]?(search|topics|media|info|tags|article|content|links)|"
            r"(search|topics|media|info|tags|article|content|links)[-\s\_]?related"
        )
        self.div_to_p_re = r"<(a|blockquote|dl|div|img|ol|p|pre|table|ul)"
        self.caption_re = "^caption$"
        self.google_re = " google "
        self.entries_re = "^[^entry-]more.*$"
        self.facebook_re = "[^-]facebook"
        self.facebook_broadcasting_re = "facebook-broadcasting"
        self.twitter_re = "[^-]twitter|twitter-tweet"
        self.contains_article = (
            './/article|.//*[@id="article"]|.//*[contains(@itemprop,"articleBody")]'
        )

    def clean(self, doc_to_clean):
        """Remove chunks of the DOM as specified"""
        doc_to_clean = self.clean_body_classes(doc_to_clean)
        doc_to_clean = self.clean_article_tags(doc_to_clean)
        doc_to_clean = self.clean_em_tags(doc_to_clean)
        doc_to_clean = self.remove_drop_caps(doc_to_clean)
        doc_to_clean = self.remove_scripts_styles(doc_to_clean)
        doc_to_clean = self.clean_bad_tags(doc_to_clean)

        # Remove image captions
        doc_to_clean = self.remove_nodes_regex(doc_to_clean, self.caption_re)
        doc_to_clean = self.clean_caption_tags(doc_to_clean)

        doc_to_clean = self.remove_nodes_regex(doc_to_clean, self.google_re)
        doc_to_clean = self.remove_nodes_regex(doc_to_clean, self.entries_re)

        # Remove social media cards
        doc_to_clean = self.remove_nodes_regex(doc_to_clean, self.facebook_re)
        doc_to_clean = self.remove_nodes_regex(doc_to_clean, self.twitter_re)
        doc_to_clean = self.remove_nodes_regex(
            doc_to_clean, self.facebook_broadcasting_re
        )
        # Remove "related" sections
        doc_to_clean = self.remove_nodes_regex(
            doc_to_clean, self.remove_nodes_related_re
        )

        # Remove spans inside of paragraphs
        doc_to_clean = self.clean_para_spans(doc_to_clean)

        doc_to_clean = self.reduce_article(doc_to_clean)

        return doc_to_clean

    def clean_whitespace(self, text: str) -> str:
        """Remove tabs, whitespace lines from text
        add double newlines to paragraphs
        """
        text = text.replace("\t", "")
        text = re.sub(r"(?<!\n)\n(?!\n)", "\n\n", text)
        text = re.sub(r"^\s+$", "", text)
        return text

    def clean_body_classes(self, doc):
        """Removes the `class` attribute from the <body> tag because
        if there is a bad match, the entire DOM will be empty!
        """
        elements = parsers.get_tags(doc, tag="body")
        if elements:
            # Remove attribute
            elements[0].attrib.pop("class", None)
        return doc

    def clean_article_tags(self, doc):
        articles = parsers.get_tags(doc, tag="article")

        # Remove this attribute from every <article> tag
        for article in articles:
            for attr in ["id", "name", "class"]:
                article.attrib.pop(attr, None)
        return doc

    def clean_em_tags(self, doc):
        ems = parsers.get_tags(doc, tag="em")
        for node in ems:
            images = parsers.get_tags(node, tag="img")
            if len(images) == 0:
                parsers.drop_tags(node)
        return doc

    def clean_caption_tags(self, doc):
        captions = parsers.get_tags(doc, tag="figcaption")
        parsers.remove(captions)

        captions = parsers.get_tags(doc, attribs={"itemprop": "caption"})
        parsers.remove(captions)

        captions = parsers.get_tags(doc, attribs={"class": "instagram-media"})
        parsers.remove(captions)

        captions = parsers.get_tags(doc, attribs={"class": "image-caption"})
        parsers.remove(captions)

        captions = parsers.get_tags(
            doc, attribs={"class": "caption"}, attribs_match="substring"
        )
        captions = [c for c in captions if c.tag in ["div", "span", "header"]]
        parsers.remove(captions)

        return doc

    def remove_drop_caps(self, doc):
        items = parsers.get_tags(
            doc, "span", {"class": "dropcap"}, attribs_match="word"
        )
        items += parsers.get_tags(
            doc, "span", {"class": "drop_cap"}, attribs_match="word"
        )

        parsers.drop_tags(items)
        return doc

    def remove_scripts_styles(self, doc):
        # remove scripts
        scripts = parsers.get_tags(doc, tag="script")
        for item in scripts:
            parsers.remove(item)
        # remove styles
        styles = parsers.get_tags(doc, tag="style")
        for item in styles:
            parsers.remove(item)
        # remove comments <--! like this one -->
        comments = doc.xpath("//comment()")

        parsers.remove(comments)

        return doc

    def clean_bad_tags(self, doc):
        # bad ids
        naughty_list = parsers.get_tags_regex(doc, attribs={"id": self.remove_nodes_re})
        for node in naughty_list:
            if not node.xpath(self.contains_article):
                parsers.remove(node)
        # class
        naughty_list = parsers.get_tags_regex(
            doc, attribs={"class": self.remove_nodes_re}
        )
        for node in naughty_list:
            if not node.xpath(self.contains_article):
                parsers.remove(node)
        # name
        naughty_list = parsers.get_tags_regex(
            doc, attribs={"name": self.remove_nodes_re}
        )
        parsers.remove(naughty_list)

        # Navigation, menus, headers, footers, etc.
        bad_tags = ["aside", "nav", "noscript", "menu"]
        naughty_list = parsers.get_elements_by_tagslist(doc, bad_tags)
        parsers.remove(naughty_list)

        return doc

    def remove_nodes_regex(self, doc, pattern):
        naughty_list = parsers.get_tags_regex(doc, attribs={"id": pattern})
        naughty_list += parsers.get_tags_regex(doc, attribs={"class": pattern})

        parsers.remove(naughty_list)

        return doc

    def clean_para_spans(self, doc):
        spans = doc.xpath(".//p/span")
        parsers.drop_tags(spans)
        return doc

    def reduce_article(self, doc):
        body_tag = parsers.get_tags(doc, tag="body")
        if not body_tag:
            return doc

        for item in body_tag[0].iter():
            if item.tag not in [
                "p",
                "br",
                "img",
                "h1",
                "h2",
                "h3",
                "h4",
                "h5",
                "h6",
                "ul",
                "body",
                "article",
                "section",
            ]:
                if item.text is None and item.tail is None:
                    item.drop_tag()

        return doc



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)
"""Module providing the Article class for newspaper. The Article class
abstracts the concept of a news article, providing methods and properties
to download, parse and analyze said article.
"""

from datetime import datetime
import json
import logging
import copy
from typing import Any, Dict, List, Optional, Set, Union
from urllib.parse import urlparse
import lxml

import requests

from newspaper.exceptions import ArticleException

from . import network
from . import nlp
from . import settings
from . import urls

from .cleaners import DocumentCleaner
from .configuration import Configuration
import newspaper.parsers as parsers
from .extractors import ContentExtractor
from .outputformatters import OutputFormatter
from .utils import (
    URLHelper,
    RawHelper,
    get_available_languages,
    extract_meta_refresh,
)

log = logging.getLogger(__name__)

available_requests_params = [
    "headers",
    "cookies",
    "auth",
    "timeout",
    "allow_redirects",
    "proxies",
    "verify",
    "cert",
]


class ArticleDownloadState:
    """Download state for the Article object."""

    NOT_STARTED = 0
    FAILED_RESPONSE = 1
    SUCCESS = 2


class Article:
    """Article abstraction for newspaper.

    This object fetches and holds information for a single article.
    In order to download the article, call `download()`. Then call `parse()`
    to extract the information.

    Attributes:
        config (Configuration): the active configuration for this article instance.
            You can use different settings for any article instance.
        extractor (ContentExtractor): Content parsing object.
        source_url (str): URL to the main page of the news source which
            owns this article
        url (str): The article link. This was used to download the current
            article. In case of a redirect(through meta refresh or read
            more link), this will be different from the original url.
        original_url (str): The original url of the article. This is the url
            that was passed to the constructor. It will not change in
            case of a redirect.
        title (str): Parset title of the article. It can be forced/overridden
            by providing a title in the constructor.
        read_more_link (str): An xpath selector for the link to the full
            article. make sure that the selector works for all casese, not
            only for one specific article. If needed, you can use several xpath
            selectors separated by ``|``.
        top_image (str): The top image url of the article. It will try to guess
            the best fit for a main image from the images found in the article.
        meta_img (str): Image url provided by metadata
        images (List[str]): List of all image urls in the current article
        movies (List[str]): List of video links in the article body
        text (str): a parsed version of the article body. It will be truncated
            to the first `config.max_text` characters.
        text_cleaned (str): a parsed version of the clean_top_node content.
            It will be truncated to the first `config.max_text` characters.
        keywords (List[str]): An inferred list of keywords for this article.
            This will be generated by the nlp method. It will be truncated to
            the first `config.max_keywords` keywords.
        keyword_scores (Dict[str, float]): A dictionary of keywords and their
            scores.
        meta_keywords (List[str]):  A list of keywords provided by the meta data.
            It will be truncated to the first `config.max_keywords` keywords.
        tags (Set[str]): Extracted tag list from the article body
        authors (List[str]): The author list parsed from the article. It will
            be truncated to the first `config.max_authors` authors.
        publish_date (str): The parsed publishing date from the article. If no
            valid date is found, it will be an empty string.
        summary (str): The summarization of the article as generated by the nlp
            method. It will be truncated to the first `config.max_summary_sent`
            sentences.
        html (str): The raw html of the article page.
        article_html (str): The raw html of the article body.
        is_parsed (bool): True if parse() has been called.
        download_state (int): AticleDownloadState.SUCCESS if `download()` was
            successful, ArticleDownloadState.FAILED_RESPONSE if `download()` failed,
            `ArticleDownloadState.NOT_STARTED` if `download()` was not called.
        download_exception_msg (str): The exception message if download() failed.
        history (List[str]): Redirection history from the requests.get call.
        meta_description (str): The description extracted from the meta data.
        meta_lang (str): The language extracted from the meta data.
            If config.language is not set, this value will be used
            to parse the article instead of the config.language value.
        meta_favicon (str): Website's favicon url extracted from the meta data.
        meta_site_name (str): Website's name extracted from the meta data.
        meta_data (Dict[str, str]): additional meta data extracted from
            the meta tags.
        canonical_link (str): Canonical URL for the article extracted from the metadata
        top_node (lxml.html.HtmlElement): Top node of the original DOM tree.
            It contains the text nodes for the detected article body. This node
            is on the doc DOM tree.
        clean_top_node (lxml.html.HtmlElement): Top node for the article on the
            cleaned version of the DOM. This node is _not_ in the doc DOM tree.
        doc (lxml.html.HtmlElement): the full DOM of the downloaded html. It is
            the original DOM tree.
        clean_doc (lxml.html.HtmlElement): a cleaned version of the DOM tree
        additional_data (Dict[Any, Any]): A property dict for users to store
            custom data.
        link_hash (str): a unique hash for the url of this article. It is salted
            with the timestamp of the download.
    """

    def __init__(
        self,
        url: str,
        title: str = "",
        source_url: str = "",
        read_more_link: str = "",
        config: Optional[Configuration] = None,
        **kwargs: Dict[str, Any],
    ):
        """Constructs the article class. Will not download or parse the article

        Args:
            url (str): The input url to parse. Can be a URL or a file path.
            title (str, optional): Default title if none can be
                extracted from the webpage. Defaults to "".
            source_url (str, optional): URL of the main website that
                originates the article.
                If left empty, it will be inferred from the url. Defaults to "".
            read_more_link (str, optional): A xpath selector for the link to the
                full article, in case there is a 'preview' with a read-more
                button that leads to another url (even on another domain).
                make sure that the selector works for all cases,
                not only for one specific article. If needed, you can use
                several xpath selectors separated by `|`. Defaults to "".
            config (Configuration, optional): Configuration settings for
            this article's download/parsing/nlp. If left empty, it will
            use the default settingsDefaults to None.

        Keyword Args:
            **kwargs: Any Configuration class propriety can be overwritten
                    through init keyword  params.
                    Additionally, you can specify any of the following
                    requests parameters:
                    headers, cookies, auth, timeout, allow_redirects,
                    proxies, verify, cert

        Raises:
            ArticleException: Error parsing and preparing the article
        """
        if isinstance(title, Configuration) or isinstance(source_url, Configuration):
            raise ArticleException(
                "Configuration object being passed incorrectly as title or "
                "source_url! Please verify `Article`s __init__() fn."
            )

        self.config: Configuration = config or Configuration()
        # Set requests parameters. These are passed directly to requests.get
        for k in available_requests_params:
            if k in kwargs:
                self.config.requests_params[k] = kwargs[k]
                del kwargs[k]
        self.config.update(**kwargs)

        self.extractor = ContentExtractor(self.config)

        if source_url == "":
            scheme = urls.get_scheme(url)
            if scheme is None:
                scheme = "http"
            source_url = scheme + "://" + urls.get_domain(url)

        if source_url is None or source_url == "":
            raise ArticleException("input url bad format")

        # URL to the main page of the news source which owns this article
        self.source_url = source_url

        self.url = urls.prepare_url(url, self.source_url)

        # In case of follow read more link, we need to keep the original url
        self.original_url = self.url

        self._title = title
        self.title = title

        # An xpath that allows to find the link to the full article
        self.read_more_link = read_more_link

        # URL of the "best image" to represent this article
        self.top_image = ""

        # stores image provided by metadata
        self.meta_img = ""

        # All image urls in this article
        self.images: List[str] = []

        # All videos in this article: youtube, vimeo, etc
        self.movies: List[str] = []

        # Body text from this article
        self._text = ""
        self.text_cleaned = ""

        # `keywords` are extracted via nlp() from the body text
        self.keywords: List[str] = []

        # `keyword_scores` a dictionary of keywords and their scores
        self.keyword_scores: Dict[str, float] = {}

        # `meta_keywords` are extracted via parse() from <meta> tags
        self.meta_keywords: List[str] = []

        # `tags` are also extracted via parse() from <meta> tags
        self.tags: Set[str] = set()

        # List of authors who have published the article, via parse()
        self.authors: List[str] = []

        self.publish_date: Optional[datetime] = None

        # Summary generated from the article's body txt
        self._summary = ""

        # This article's unchanged and raw HTML
        self._html = ""

        # The HTML of this article's main node (most important part)
        self.article_html = ""

        # Keep state for downloads and parsing
        self.is_parsed = False
        self.download_state = ArticleDownloadState.NOT_STARTED
        self.download_exception_msg: Optional[str] = None

        # Redirection history from the requests.get call
        self.history: Optional[List[str]] = []

        # Meta description field in the HTML source
        self.meta_description = ""

        # Meta language field in HTML source
        self.meta_lang = ""

        # Meta favicon field in HTML source
        self.meta_favicon = ""

        # Meta site_name field in HTML source
        self.meta_site_name = ""

        # Meta tags contain a lot of structured data, e.g. OpenGraph
        self.meta_data: Dict[str, str] = {}

        # The canonical link of this article if found in the meta data
        self.canonical_link = ""

        # Holds the top element of the DOM that we determine is a candidate
        # for the main body of the article
        self.top_node: Optional[lxml.html.Element] = None

        # A deepcopied clone of the above object before heavy parsing
        # operations, useful for users to query data in the
        # "most important part of the page"
        self.clean_top_node: Optional[lxml.html.Element] = None

        # The top node complemented with siblings (off-tree)
        self._top_node_complemented: Optional[lxml.html.Element] = None

        # lxml DOM object generated from HTML
        self.doc: Optional[lxml.html.Element] = None

        # A deepcopied clone of the above object before undergoing heavy
        # cleaning operations, serves as an API if users need to query the DOM
        self.clean_doc: Optional[lxml.html.Element] = None

        # A property dict for users to store custom data.
        self.additional_data: Dict[Any, Any] = {}

        self.link_hash: Optional[str] = None

    def build(self):
        """Build a lone article from a URL independent of the source (newspaper).
        Don't normally call this method b/c it's good to multithread articles
        on a source (newspaper) level.
        Calls download(), parse(), and nlp() in succession.
        """
        self.download()
        self.parse()
        self.nlp()

    def _parse_scheme_file(self, path):
        try:
            with open(path, "r", encoding="utf-8") as fin:
                return fin.read()
        except OSError as e:
            self.download_state = ArticleDownloadState.FAILED_RESPONSE
            self.download_exception_msg = e.strerror
            return None

    def _parse_scheme_http(self, url: Optional[str] = None):
        try:
            # We do not use get_html() here because we want to be able to
            # detect protection in the response regardless of the status code
            html, status_code, history = network.get_html_status(
                url or self.url, self.config
            )
            self.history = [r.url for r in history]
            if status_code >= 400:
                self.download_state = ArticleDownloadState.FAILED_RESPONSE
                protection = self._detect_protection(html)
                if protection:
                    self.download_exception_msg = (
                        f"Website protected with {protection}, url: {url}"
                    )
                else:
                    self.download_exception_msg = (
                        f"Status code {status_code} for url {url}"
                    )
                return None
        except requests.exceptions.RequestException as e:
            self.download_state = ArticleDownloadState.FAILED_RESPONSE
            self.download_exception_msg = str(e)
            return None

        return html

    def _detect_protection(self, html):
        if "cloudflare" in html:
            return "Cloudflare"
        if "/cdn-cgi/challenge-platform/h/b/orchestrate/chl_page" in html:
            return "Cloudflare"
        if "cloud-flare" in html:
            return "Cloudflare"
        if "CloudFront" in html:
            return "CloudFront"
        if "perimeterx" in html:
            return "PerimeterX"

        return None

    def download(
        self,
        input_html: Optional[str] = None,
        title: Optional[str] = None,
        recursion_counter: int = 0,
        ignore_read_more: bool = False,
    ) -> "Article":
        """Downloads the link's HTML content, don't use if you are batch async
        downloading articles

        Args:
            input_html (str, optional): A cached version of the article to parse.
                It will load the html from this string without attempting to access
                the article url. If you have a read_more_link xpath
                set up in the constructor, and do not set ignore_read_more to true,
                it will attempt to follow the found read_more link (if any).
                Defaults to None.
            title (str, optional): Force an article title. Defaults to None.
            recursion_counter (int, optional): Used to prevent infinite recursions
            due to meta_refresh. Defaults to 0.
            ignore_read_more (bool, optional): If true, the download process will
            ignore any kind of "read_more" xpath set up in the constructor.
            Defaults to False.
        Returns:
            Article: self
        """

        if input_html is None:
            parsed_url = urlparse(self.url)
            if parsed_url.scheme == "file":
                html = self._parse_scheme_file(parsed_url.path)
            else:
                html = self._parse_scheme_http()
            if html is None:
                log.debug(
                    "Download failed on URL %s because of %s",
                    self.url,
                    self.download_exception_msg,
                )
                return self
        else:
            html = input_html

        if self.config.follow_meta_refresh:
            meta_refresh_url = extract_meta_refresh(html)
            if meta_refresh_url and recursion_counter < 1:
                return self.download(
                    input_html=network.get_html(meta_refresh_url),
                    recursion_counter=recursion_counter + 1,
                )

        if not ignore_read_more and self.read_more_link:
            doc = parsers.fromstring(html)
            for read_more_node in doc.xpath(self.read_more_link):
                # TODO: add check for onclick redirections. need some examples
                if read_more_node.get("href"):
                    new_url = read_more_node.get("href")
                    log.info(
                        "After downloading %s, found read more link: %s",
                        self.url,
                        new_url,
                    )
                    new_url = urls.prepare_url(new_url, self.url)
                    html_ = self._parse_scheme_http(new_url)
                    if html_ is not None:
                        html = html_
                        self.url = new_url
                        log.info(
                            "Downloaded read more link: %s and updated url to %s",
                            new_url,
                            self.url,
                        )
                    else:
                        log.info(
                            "Failed to download read more link: %s, leaving original"
                            " content in place",
                            new_url,
                        )
                    break

        self.html = html
        self.title = title

        return self

    def parse(self) -> "Article":
        """Parse the previously downloaded article.
        If `download()` wasn't called, it will raise
        a `ArticleException` exception.
        Populates the article properties such as:
        ``title``, ``authors``, ``publish_date``,
        ``text``, ``top_image``, etc.
        Returns:
            Article: self
        """
        self.throw_if_not_downloaded_verbose()

        self.doc = parsers.fromstring(self.html)
        self.clean_doc = copy.deepcopy(self.doc)

        if self.doc is None:
            # `parse` call failed, return nothing
            self.is_parsed = True
            return self

        # TODO: Fix this, sync in our fix_url() method
        parse_candidate = self.get_parse_candidate()
        self.link_hash = parse_candidate.link_hash  # MD5

        document_cleaner = DocumentCleaner(self.config)
        output_formatter = OutputFormatter(self.config)

        title = self.extractor.get_title(self.clean_doc)
        self.title = title

        authors = self.extractor.get_authors(self.clean_doc)
        self.authors = authors[: self.config.max_authors]

        metadata = self.extractor.get_metadata(self.url, self.clean_doc)
        if metadata["language"] in get_available_languages():
            self.meta_lang = metadata["language"]

            if self.config.use_meta_language:
                self.extractor.update_language(self.meta_lang)

        self.meta_site_name = metadata["site_name"]
        self.meta_description = metadata["description"]
        self.canonical_link = metadata["canonical_link"]
        self.meta_keywords = metadata["keywords"]
        self.tags = metadata["tags"]
        self.meta_data = metadata["data"]

        self.publish_date = self.extractor.get_publishing_date(self.url, self.clean_doc)

        # Before any computations on the body, clean DOM object
        self.clean_doc = document_cleaner.clean(self.clean_doc)

        # Top node in the original documentDOM
        self.top_node = self.extractor.calculate_best_node(self.doc)
        # Off-tree Node containing the top node and any relevant siblings
        self._top_node_complemented = self.extractor.top_node_complemented

        # Top node in the cleaned version of the DOM
        self.clean_top_node = self.extractor.calculate_best_node(self.clean_doc)

        self.set_movies(self.extractor.get_videos(self.doc, self.top_node))

        if self.top_node is not None:
            self._top_node_complemented = document_cleaner.clean(
                self._top_node_complemented
            )
            text, article_html = output_formatter.get_formatted(
                self._top_node_complemented, title
            )
            self.article_html = article_html
            self.text = text

            text, _ = output_formatter.get_formatted(self.clean_top_node, title)
            self.text_cleaned = text[: self.config.max_text] if text else ""

        self.fetch_images()

        self.is_parsed = True
        return self

    def fetch_images(self):
        """Fetch top image, meta image and image list from
        current cleaned_doc. Will set the attributes: meta_img,
        top_image, images, meta_favicon
        """
        # TODO: check weather doc or clean doc is better
        # TODO: rewrite set_reddit_top_img. I removed it for now
        self.extractor.parse_images(self.url, self.clean_doc, self.clean_top_node)

        self.meta_img = self.extractor.image_extractor.meta_image
        self.top_image = self.extractor.image_extractor.top_image
        self.images = self.extractor.image_extractor.images
        self.meta_favicon = self.extractor.image_extractor.favicon

    def is_valid_url(self):
        """Performs a check on the url of this link to determine if article
        is a real news article or not
        """
        return urls.valid_url(self.url)

    def is_valid_body(self):
        """If the article's body text is long enough to meet
        standard article requirements, keep the article
        """
        if not self.is_parsed:
            raise ArticleException(
                "must parse article before checking                                    "
                " if it's body is valid!"
            )
        meta_type = self.extractor.metadata_extractor.meta_data["type"]
        wordcount = self.text.split(" ")
        sentcount = self.text.split(".")

        if meta_type == "article" and len(wordcount) > (self.config.min_word_count):
            log.debug("%s verified for article and wc", self.url)
            return True

        if not self.is_media_news() and not self.text:
            log.debug("%s caught for no media no text", self.url)
            return False

        if self.title is None or len(self.title.split(" ")) < 2:
            log.debug("%s caught for bad title", self.url)
            return False

        if len(wordcount) < self.config.min_word_count:
            log.debug("%s caught for word cnt", self.url)
            return False

        if len(sentcount) < self.config.min_sent_count:
            log.debug("%s caught for sent cnt", self.url)
            return False

        if self.html is None or self.html == "":
            log.debug("%s caught for no html", self.url)
            return False

        log.debug("%s verified for default true", self.url)
        return True

    def is_media_news(self):
        """If the article is related heavily to media:
        gallery, video, big pictures, etc
        """
        safe_urls = [
            "/video",
            "/slide",
            "/gallery",
            "/powerpoint",
            "/fashion",
            "/glamour",
            "/cloth",
        ]
        for s in safe_urls:
            if s in self.url:
                return True
        return False

    def nlp(self):
        """Method expects `download()` and `parse()` to have been run.
        It will perform the keyword extraction and summarization"""
        self.throw_if_not_downloaded_verbose()
        self.throw_if_not_parsed_verbose()

        nlp.load_stopwords(self.config.language)
        keywords = nlp.keywords(self.text, self.config.max_keywords)
        for k, v in nlp.keywords(self.title, self.config.max_keywords).items():
            if k in keywords:
                keywords[k] += v
                keywords[k] /= 2
            else:
                keywords[k] = v

        keywords = sorted(list(keywords.items()), key=lambda x: x[1], reverse=True)
        keywords = keywords[: self.config.max_keywords]

        self.keywords = [x[0] for x in keywords]  # remove score
        self.keyword_scores = dict(keywords)

        max_sents = self.config.max_summary_sent

        summary_sents = nlp.summarize(
            title=self.title, text=self.text, max_sents=max_sents
        )
        self.summary = "\n".join(summary_sents)

    def get_parse_candidate(self):
        """A parse candidate is a wrapper object holding a link hash of this
        article and a final_url of the article
        """
        if self.html:
            return RawHelper.get_parsing_candidate(self.url, self.html)
        return URLHelper.get_parsing_candidate(self.url)

        # os.remove(path)

    @property
    def title(self) -> str:
        return self._title

    @title.setter
    def title(self, value: str):
        self._title = value[: self.config.max_title] if value else ""

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, value: str):
        self._text = value[: self.config.max_text] if value else ""

    @property
    def html(self) -> str:
        return self._html

    @html.setter
    def html(self, value: str):
        self.download_state = ArticleDownloadState.SUCCESS
        if value:
            if isinstance(value, bytes):
                value = parsers.get_unicode_html(value)
            self._html = value
        else:
            self._html = ""

    @property
    def imgs(self) -> List[str]:
        """Same as images

        Returns:
            List[str]: list of image urls
        """
        # Seems to be some legacy api,
        return self.images

    @property
    def top_img(self) -> str:
        """Same as top_image

        Returns:
            str: top_image
        """
        # Seems to be some legacy api,
        return self.top_image

    @property
    def summary(self) -> str:
        return self._summary

    @summary.setter
    def summary(self, value: str):
        self._summary = value[: self.config.max_summary] if value else ""

    def set_movies(self, movie_objects):
        """Trim video objects into just urls"""
        movie_urls = [o.src for o in movie_objects if o and o.src]
        self.movies = movie_urls

    def throw_if_not_downloaded_verbose(self):
        """Parse ArticleDownloadState -> log readable status
        -> maybe throw ArticleException
        """
        if self.download_state == ArticleDownloadState.NOT_STARTED:
            raise ArticleException("You must `download()` an article first!")
        elif self.download_state == ArticleDownloadState.FAILED_RESPONSE:
            raise ArticleException(
                "Article `download()` failed with %s on URL %s"
                % (self.download_exception_msg, self.url)
            )

    def throw_if_not_parsed_verbose(self):
        """Parse `is_parsed` status -> log readable status
        -> maybe throw ArticleException
        """
        if not self.is_parsed:
            raise ArticleException("You must `parse()` an article first!")

    def to_json(self, as_string: Optional[bool] = True) -> Union[str, Dict]:
        """Create a json string from the article data. It will include the most
        important attributes such as title, text, authors, publish_date, etc.
        Must be called after `parse()`

        Arguments:
            as_string (bool, optional): If True, it will return a json string.
                If False, it will return a json object. Defaults to True.

        Returns:
            str: the json string version of an parsed article.
        """

        self.throw_if_not_parsed_verbose()

        article_dict = {}

        for metadata in settings.article_json_fields:
            article_dict[metadata] = getattr(
                self, metadata, getattr(self.config, metadata, None)
            )
            if isinstance(article_dict[metadata], datetime):
                article_dict[metadata] = article_dict[metadata].isoformat()
        if as_string:
            return json.dumps(article_dict, indent=4, ensure_ascii=False)
        else:
            return article_dict



# -*- coding: utf-8 -*-

# Copyright (c) [2023] [Andrei Paraschiv]
#
# This file is part of [Newspaper4k].
#   https://github.com/AndyTheFactory/newspaper4k
#
# [Newspaper4k] includes code from the original project,
# [Newspaper3k], which is licensed under [MIT].
#
# I would like to express gratitude to the creator of [Newspaper3k],
# Lucas Ou-Yang (codelucas) for their valuable work.
# You can find the original project here: https://github.com/codelucas/newspaper


from typing import Optional
from .api import (
    build,
    build_article,
    fulltext,
    hot,
    languages,
    popular_urls,
    Configuration as Config,
)
from .article import Article
from .source import Source
from .version import __version__
import logging
from logging import NullHandler
from .exceptions import ArticleBinaryDataException, ArticleException

# Set default logging handler to avoid "No handler found" warnings.
logging.getLogger(__name__).addHandler(NullHandler())


def article(url: str, language: Optional[str] = "en", **kwargs) -> Article:
    """Shortcut function to fetch and parse a newspaper article from a URL.

    Args:
        url (str): The URL of the article to download and parse.
        language (str): The language of the article to download and parse.
        input_html (str): The HTML of the article to parse. This
            is used for pre-downloaded articles. If this is set,
            then there will be no download requests made.
        kwargs: Any other keyword arguments to pass to the Article constructor.

    Returns:
        Article: The article downloaded and parsed.

    Raises:
        ArticleException: If the article could not be downloaded or parsed.
    """
    if "input_html" in kwargs:
        input_html = kwargs["input_html"]
        del kwargs["input_html"]
    else:
        input_html = None
    a = Article(url, language=language, **kwargs)
    a.download(input_html=input_html)
    a.parse()
    return a


__all__ = [
    "build",
    "build_article",
    "article",
    "fulltext",
    "hot",
    "languages",
    "popular_urls",
    "Config",
    "Article",
    "ArticleException",
    "ArticleBinaryDataException",
    "Source",
    "__version__",
]



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

"""
Stopword extraction and stopword classes.
"""

from dataclasses import dataclass, field
from pathlib import Path
import re
import string
from typing import Dict, List

from newspaper import settings


def innerTrim(value):
    if isinstance(value, str):
        # remove tab and white space
        value = re.sub(r"[\s\t]+", " ", value)
        value = "".join(value.splitlines())
        return value.strip()
    return ""


@dataclass
class WordStats:
    """Holds the number of stop words and total words in an article"""

    stop_word_count: int = 0
    word_count: int = 0
    stop_words: List[str] = field(default_factory=list)


class StopWords:
    TRANS_TABLE = str.maketrans("", "")
    _cached_stop_words: Dict[str, str] = {}

    def __init__(self, language="en"):
        if language not in self._cached_stop_words:
            stopwordsFile = Path(settings.STOPWORDS_DIR) / f"stopwords-{language}.txt"
            if not stopwordsFile.exists():
                raise FileNotFoundError(
                    f"Stopwords file for language {language} not found! Make sure that "
                    "the language is supported (see `newspaper.languages()`)"
                )
            with open(stopwordsFile, "r", encoding="utf-8") as f:
                self._cached_stop_words[language] = set(f.read().splitlines())

        self.STOP_WORDS = self._cached_stop_words[language]

    def remove_punctuation(self, content):
        # code taken form
        # http://stackoverflow.com/questions/265960/best-way-to-strip-punctuation-from-a-string-in-python
        content_is_unicode = isinstance(content, str)
        if content_is_unicode:
            content = content.encode("utf-8")
        trans_table = {ord(c): None for c in string.punctuation}
        stripped_input = content.decode("utf-8").translate(trans_table)

        return stripped_input

    def candidate_words(self, stripped_input):
        return stripped_input.split(" ")

    def get_stopword_count(self, content):
        if not content:
            return WordStats()
        ws = WordStats()
        stripped_input = self.remove_punctuation(content)
        candidate_words = self.candidate_words(stripped_input.lower())
        overlapping_stopwords = []
        c = 0
        for w in candidate_words:
            c += 1
            if w in self.STOP_WORDS:
                overlapping_stopwords.append(w)

        ws.word_count = c
        ws.stop_word_count = len(overlapping_stopwords)
        ws.stop_words = overlapping_stopwords
        return ws


class StopWordsChinese(StopWords):
    """Chinese segmentation"""

    def __init__(self, language="zh"):
        super(StopWordsChinese, self).__init__(language="zh")

    def candidate_words(self, stripped_input):
        # jieba builds a tree that takes a while. avoid building
        # this tree if we don't use the chinese language
        import jieba

        return jieba.cut(stripped_input, cut_all=True)


class StopWordsArabic(StopWords):
    """Arabic segmentation"""

    def __init__(self, language="ar"):
        # force ar language code
        super(StopWordsArabic, self).__init__(language="ar")

    def remove_punctuation(self, content):
        return content

    def candidate_words(self, stripped_input):
        import nltk

        s = nltk.stem.isri.ISRIStemmer()
        words = []
        for word in nltk.tokenize.wordpunct_tokenize(stripped_input):
            words.append(s.stem(word))
        return words


class StopWordsKorean(StopWords):
    """Korean segmentation"""

    def __init__(self, language="ko"):
        super(StopWordsKorean, self).__init__(language="ko")

    def get_stopword_count(self, content):
        if not content:
            return WordStats()
        ws = WordStats()
        stripped_input = self.remove_punctuation(content)
        candidate_words = self.candidate_words(stripped_input)
        overlapping_stopwords = []
        c = 0
        for w in candidate_words:
            c += 1
            for s in self.STOP_WORDS:
                if w.endswith(s):
                    overlapping_stopwords.append(w)

        ws.word_count = c
        ws.stop_word_count = len(overlapping_stopwords)
        ws.stop_words = overlapping_stopwords
        return ws


class StopWordsHindi(StopWords):
    """Hindi segmentation"""

    def __init__(self, language="hi"):
        super(StopWordsHindi, self).__init__(language="hi")

    def get_stopword_count(self, content):
        if not content:
            return WordStats()
        ws = WordStats()
        stripped_input = self.remove_punctuation(content)
        candidate_words = self.candidate_words(stripped_input)
        overlapping_stopwords = []
        c = 0
        for w in candidate_words:
            c += 1
            for stop_word in self.STOP_WORDS:
                overlapping_stopwords.append(stop_word)

        ws.word_count = c
        ws.stop_word_count = len(overlapping_stopwords)
        ws.stop_words = overlapping_stopwords
        return ws


class StopWordsJapanese(StopWords):
    """Japanese segmentation"""

    def __init__(self, language="ja"):
        super(StopWordsJapanese, self).__init__(language="ja")

    def candidate_words(self, stripped_input):
        import tinysegmenter

        segmenter = tinysegmenter.TinySegmenter()
        tokens = segmenter.tokenize(stripped_input)
        return tokens


class StopWordsThai(StopWords):
    """Thai segmentation"""

    def __init__(self, language="th"):
        super(StopWordsThai, self).__init__(language="th")

    def candidate_words(self, stripped_input):
        import pythainlp

        tokens = pythainlp.word_tokenize(stripped_input)
        return tokens



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

"""
Functions needed for the NLP analysis of articles.
"""


import os
import re
import math
from pathlib import Path
from collections import Counter
from typing import List, Set

from . import settings

stopwords: Set[str] = set()


def load_stopwords(language):
    """
    Loads language-specific stopwords for keyword selection
    """
    # stopwords for nlp in English are not the regular stopwords
    # to pass the tests
    # can be changed with the tests

    stopwordsFile = Path(settings.STOPWORDS_DIR) / f"stopwords-{language}.txt"
    if not stopwordsFile.exists():
        raise ValueError(
            f"Language {language} is not supported "
            "(or make sure the stopwords file is present in "
            "{settings.STOPWORDS_DIR}), please use one of "
            "the following: {settings.languages}"
        )
    with open(stopwordsFile, "r", encoding="utf-8") as f:
        stopwords.update(set([w.strip() for w in f.readlines()]))


def keywords(text, max_keywords=None):
    """Get the top 10 keywords and their frequency scores ignores blacklisted
    words in stopwords, counts the number of occurrences of each word, and
    sorts them in reverse natural order (so descending) by number of
    occurrences.
    """
    # TODO: parametrable number of keywords
    text = split_words(text)
    if not text:
        return dict()
    # of words before removing blacklist words
    num_words = len(text)
    text = [x for x in text if x not in stopwords]
    freq = {}
    for word in text:
        if word in freq:
            freq[word] += 1
        else:
            freq[word] = 1

    keywords_ = sorted(freq.items(), key=lambda x: (x[1], x[0]), reverse=True)
    if max_keywords:
        keywords_ = keywords_[:max_keywords]
    keywords_ = dict((x, y) for x, y in keywords_)

    for k in keywords_:
        articleScore = keywords_[k] * 1.0 / max(num_words, 1)
        keywords_[k] = articleScore * 1.5 + 1
    return dict(keywords_)


def summarize(url="", title="", text="", max_sents=5):
    if not text or not title or max_sents <= 0:
        return []

    summaries = []
    sentences = split_sentences(text)
    keys = keywords(text, settings.SUMMARIZE_KEYWORD_COUNT)
    titleWords = split_words(title)

    # Score sentences, and use the top 5 or max_sents sentences
    ranks = score(sentences, titleWords, keys).most_common(max_sents)
    for rank in ranks:
        summaries.append(rank[0])
    summaries.sort(key=lambda summary: summary[0])
    return [summary[1] for summary in summaries]


def score(sentences, titleWords, keywords):
    """Score sentences based on different features"""
    senSize = len(sentences)
    ranks = Counter()
    for i, s in enumerate(sentences):
        sentence = split_words(s)
        titleFeature = title_score(titleWords, sentence)
        sentenceLength = length_score(len(sentence))
        sentencePosition = sentence_position(i + 1, senSize)
        sbsFeature = sbs(sentence, keywords)
        dbsFeature = dbs(sentence, keywords)
        frequency = (sbsFeature + dbsFeature) / 2.0 * 10.0
        # Weighted average of scores from four categories
        totalScore = (
            titleFeature * 1.5
            + frequency * 2.0
            + sentenceLength * 1.0
            + sentencePosition * 1.0
        ) / 4.0
        ranks[(i, s)] = totalScore
    return ranks


def sbs(words, keywords):
    score = 0.0
    if len(words) == 0:
        return 0
    for word in words:
        if word in keywords:
            score += keywords[word]
    return (1.0 / math.fabs(len(words)) * score) / 10.0


def dbs(words, keywords):
    if len(words) == 0:
        return 0
    summ = 0
    first = []
    second = []

    for i, word in enumerate(words):
        if word in keywords:
            score = keywords[word]
            if first == []:
                first = [i, score]
            else:
                second = first
                first = [i, score]
                dif = first[0] - second[0]
                summ += (first[1] * second[1]) / (dif**2)
    # Number of intersections
    k = len(set(keywords.keys()).intersection(set(words))) + 1
    return 1 / (k * (k + 1.0)) * summ


def split_words(text):
    """Split a string into array of words"""
    try:
        text = re.sub(r"[^\w ]", "", text)  # strip special chars
        return [x.strip(".").lower() for x in text.split()]
    except TypeError:
        return None


def split_sentences(text: str) -> List[str]:
    """Split a large string into sentences. Uses the Punkt Sentence Tokenizer
    from the nltk module to split strings into sentences.

    Args:
        text (str): input text

    Returns:
        List[str]: a list of sentences
    """
    try:
        tokenizer = split_sentences._tokenizer
    except AttributeError:
        import nltk

        nltk_data_path = os.environ.get("NLTK_DATA")
        if nltk_data_path:
            nltk.data.path.append(nltk_data_path)
        try:
            nltk.data.find("tokenizers/punkt")
        except LookupError:
            nltk.download("punkt")

        # TODO: load a language specific tokenizer
        tokenizer = nltk.data.load("tokenizers/punkt/english.pickle")
        split_sentences._tokenizer = tokenizer

    sentences = tokenizer.tokenize(text)
    sentences = [re.sub("[\n ]+", " ", x) for x in sentences if len(x) > 10]
    return sentences


def length_score(sentence_len):
    return (
        1
        - math.fabs(settings.MEAN_SENTENCE_LEN - sentence_len)
        / settings.MEAN_SENTENCE_LEN
    )


def title_score(title, sentence):
    if title:
        title = [x for x in title if x not in stopwords]
        count = 0.0
        for word in sentence:
            if word not in stopwords and word in title:
                count += 1.0
        return count / max(len(title), 1)
    else:
        return 0


def sentence_position(i, size):
    """Different sentence positions indicate different
    probability of being an important sentence.
    """
    normalized = i * 1.0 / size
    if normalized > 1.0:
        return 0
    elif normalized > 0.9:
        return 0.15
    elif normalized > 0.8:
        return 0.04
    elif normalized > 0.7:
        return 0.04
    elif normalized > 0.6:
        return 0.06
    elif normalized > 0.5:
        return 0.04
    elif normalized > 0.4:
        return 0.05
    elif normalized > 0.3:
        return 0.08
    elif normalized > 0.2:
        return 0.14
    elif normalized > 0.1:
        return 0.23
    elif normalized > 0:
        return 0.17
    else:
        return 0



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)
"""
Helper functions for handling LXML nodes and trees.
"""
from collections import deque
import json
from math import exp
import re
import logging
import string
from html import unescape
from copy import deepcopy
from typing import List, Dict, Optional, Union
import lxml.etree
import lxml.html
import lxml.html.clean

from bs4 import UnicodeDammit

from . import text as txt

log = logging.getLogger(__name__)


def drop_tags(nodes: Union[lxml.html.HtmlElement, List[lxml.html.HtmlElement]]):
    """Remove the tag(s), but not its children or text.
    The children and text are merged into the parent."""
    if not isinstance(nodes, list):
        nodes = [nodes]
    for node in nodes:
        node.drop_tag()


def get_unicode_html(html):
    if isinstance(html, str):
        return html
    if not html:
        return html
    converted = UnicodeDammit(html, is_html=True)
    if not converted.unicode_markup:
        raise ValueError(
            "Failed to detect encoding of article HTML, tried: %s"
            % ", ".join(converted.tried_encodings)
        )
    html = converted.unicode_markup
    return html


def fromstring(html):
    html = get_unicode_html(html)
    # Enclosed in a `try` to prevent bringing the entire library
    # down due to one article (out of potentially many in a `Source`)
    try:
        # lxml does not play well with <? ?> encoding tags
        if html.startswith("<?"):
            html = re.sub(r"^\<\?.*?\?\>", "", html, flags=re.DOTALL)
        return lxml.html.fromstring(html)
    except Exception:
        log.warning("fromstring() returned an invalid string: %s...", html[:20])
        return


def node_to_string(node):
    """conerts the tree under node to a string representation
    e.g. "<html><body>hello</body></html>"
    """
    # decode is needed at the end because etree.tostring
    # returns a python bytestring
    return lxml.etree.tostring(node, method="html").decode()


def get_tags_regex(
    node: lxml.html.Element,
    tag: Optional[str] = None,
    attribs: Optional[Dict[str, str]] = None,
) -> List[lxml.html.Element]:
    """Get list of elements of a certain tag with regex matching attributes

    Args:
        tag (str, optional): Tag to match. If None, it matches all
            tags. Defaults to None.
        attribs (Dict[str, str], optional): Dictionary containing
            attributes to match and the corresponding regex.
            Defaults to None. The result matches **all**
            attributes in the dictionary.

    Returns:
        List[lxml.html.Element]: Elements matching the tag and attributes
    """
    if not attribs:
        return get_tags(node, tag=tag)

    namespace = {"re": "http://exslt.org/regular-expressions"}
    sel_list = []

    for k, v in attribs.items():
        selector = f"re:test(@{k}, '{v}', 'i')"
        sel_list.append(selector)

    selector = ".//%s[%s]" % (tag or "*", " and ".join(sel_list))
    elems = node.xpath(selector, namespaces=namespace)
    return elems


def get_tags(
    node: lxml.html.Element,
    tag: Optional[str] = None,
    attribs: Optional[Dict[str, str]] = None,
    attribs_match: str = "exact",
):
    """Get list of elements of a certain tag with exact matching attributes

    Args:
        tag (str, optional): Tag to match. If None, it matches all
            tags. Defaults to None.
        attribs (Dict[str, str], optional): Dictionary containing
            attributes to match. Defaults to None. The result matches **all**
            attributes in the dictionary.
        attribs_match (str, optional): Match type. Can be "exact",
            "substring" or "word". Difference between "substring" and
            "word" is that "substring" matches any part of the attribute
            value while "word" matches attribs whose value is a
            whitespace-separated list of words, one of which is exactly
            our query string.
            Defaults to "exact".

    Returns:
        List[lxml.html.Element]: Elements matching the tag and attributes
    """
    if attribs_match not in ["exact", "substring", "word"]:
        raise ValueError("attribs_match must be one of 'exact', 'substring' or 'word'")
    if not attribs:
        selector = f".//{(tag or '*')}"
        elems = node.xpath(selector)
        return elems

    sel_list = []
    for k, v in attribs.items():
        trans = 'translate(@%s, "%s", "%s")' % (
            k,
            string.ascii_uppercase,
            string.ascii_lowercase,
        )
        if attribs_match == "exact":
            selector = '%s="%s"' % (trans, v.lower())
        elif attribs_match == "substring":
            selector = 'contains(%s, "%s")' % (trans, v.lower())
        elif attribs_match == "word":
            selector = 'contains(concat(" ", normalize-space(%s), " "), " %s ")' % (
                trans,
                v.lower(),
            )

        sel_list.append(selector)
    selector = ".//%s[%s]" % (tag or "*", " and ".join(sel_list))
    elems = node.xpath(selector)
    return elems


def get_elements_by_attribs(
    node: lxml.html.Element,
    attribs: Dict[str, str],
    attribs_match: str = "exact",
) -> List[lxml.html.Element]:
    """Get list of elements with exact matching attributes

    Args:
        attribs (Dict[str,str]): dictionary containing attributes to match.
            e.g. {"class":"foo", "id":"bar"}. The result matches **all**
            attributes in the dictionary.
        attribs_match (str, optional): Match type. Can be "exact",
            "substring" or "word". Difference between "substring" and
            "word" is that "substring" matches any part of the attribute
            value while "word" matches attribs whose value is a
            whitespace-separated list of words, one of which is exactly
            our query string.
            Defaults to "exact".

    Returns:
        List[lxml.html.Element]: Elements matching the attributes
    """
    return get_tags(node, attribs=attribs, attribs_match=attribs_match)


def get_metatags(
    node: lxml.html.Element, value: Optional[str] = None
) -> List[lxml.html.Element]:
    """Get list of meta tags with name, property **or** itemprop equal to
        `value`. If `value` is None, it returns all meta tags

    Args:
        node (lxml.html.Element): Element to search
        value (Optional[str], optional): Value to match. Defaults to None.

    Returns:
        List[lxml.html.Element]: Elements matching the value
    """
    if value is None:
        return get_tags(node, tag="meta")

    sel_list = [f"@name='{value}'", f"@property='{value}'", f"@itemprop='{value}'"]
    selector = "//meta[%s]" % " or ".join(sel_list)
    elems = node.xpath(selector)
    return elems


def get_elements_by_tagslist(node: lxml.html.Element, tag_list: List[str]):
    """Get list of elements with tag in `tag_list`

    Args:
        node (lxml.html.Element): Element to search
        tag_list (List[str]): List of tags to match

    Returns:
        List[lxml.html.Element]: Elements matching the tags
    """
    selector = " | ".join([f".//{tag}" for tag in tag_list])
    elems = node.xpath(selector)
    return elems


def create_element(tag, text=None, tail=None):
    t = lxml.html.HtmlElement()
    t.tag = tag
    t.text = text
    t.tail = tail
    return t


def remove(nodes: Union[lxml.html.HtmlElement, List[lxml.html.HtmlElement]]):
    """Remove the node(s) from the tree
    Arguments:
        nodes (Union[lxml.html.HtmlElement, List[lxml.html.HtmlElement]]):
            node or list of nodes to remove
    """
    # TODO: check if drop_tags can be used instead
    if not isinstance(nodes, list):
        nodes = [nodes]

    for node in nodes:
        parent = node.getparent()
        if parent is None:
            continue
        if node.tail:
            prev = node.getprevious()
            if prev is None:
                if not parent.text:
                    parent.text = ""
                parent.text += " " + node.tail
            else:
                if not prev.tail:
                    prev.tail = ""
                prev.tail += " " + node.tail
        node.clear()
        parent.remove(node)


def get_text(node):
    node_copy = deepcopy(node)
    lxml.etree.strip_elements(
        node_copy, lxml.etree.Comment, "script", "style", "select", "option", "textarea"
    )
    txts = list(node_copy.itertext())
    return txt.innerTrim(" ".join(txts).strip())


def get_attribute(node: lxml.html.Element, attr: str) -> Optional[str]:
    """get the unicode attribute of the node"""
    attr = node.attrib.get(attr, None)
    return unescape(attr) if attr else None


def set_attribute(node, attr, value=None):
    # Check if immutable attribute
    if not isinstance(
        node, (lxml.etree.CommentBase, lxml.etree.EntityBase, lxml.etree.PIBase)
    ):
        node.set(attr, value)


def outer_html(node):
    e0 = node
    if e0.tail:
        e0 = deepcopy(e0)
        e0.tail = None
    return node_to_string(e0)


def get_ld_json_object(node):
    """Get the JSON-LD object from the node"""
    # yoast seo structured data
    json_ld = get_tags(node, tag="script", attribs={"type": "application/ld+json"})
    res = []
    if json_ld:
        for script_tag in json_ld:
            try:
                schema_json = json.loads(script_tag.text)
            except Exception:
                continue
            if isinstance(schema_json, list):
                res.extend(schema_json)
            else:
                res.append(schema_json)

    return res


def get_node_depth(node: lxml.html.Element) -> int:
    """Get the depth of the node (how deep its children are)
    Arguments:
        node (lxml.html.Element): node to get the depth of
    Returns:
        int: depth of the node (1 for leaf)
    """
    queue = deque([(node, 1)])
    depths = [1]
    while queue:
        node, depth = queue.popleft()
        queue.extend([(child, depth + 1) for child in node.getchildren()])
        if node.getchildren():
            depths.append(depth + 1)

    return max(depths)


def get_level(node: lxml.html.Element) -> int:
    """Get the level of the node in the tree
    Arguments:
        node (lxml.html.Element): node to get the level of
    Returns:
        int: level of the node in the tree (0 for root)
    """
    root = node.getroottree()
    path = root.getpath(node)
    path = path.split("/")

    return len(path) - 1


def get_nodes_at_level(root: lxml.html.Element, level: int) -> List[lxml.html.Element]:
    """Get the nodes at a certain level in the tree
    Arguments:
        node (lxml.html.Element): node to get the level of
        level (int): level of the nodes to get
    Returns:
        List[lxml.html.Element]: list of nodes at the specified level
    """
    queue = deque([(root, 1)])

    result_nodes = []

    while queue:
        node, node_level = queue.popleft()

        if node_level == level:
            result_nodes.append(node)
        elif node_level < level:
            queue.extend([(child, node_level + 1) for child in node.getchildren()])

    return result_nodes


def is_highlink_density(e):
    """Checks the density of links within a node, if there is a high
    link to text ratio, then the text is less likely to be relevant
    """
    links = get_elements_by_tagslist(e, ["a", "button"])
    if not links:
        return False

    text = get_text(e)
    words = [word for word in text.split() if word.isalnum()]
    if not words:
        return len(links) > 0

    total_words = len(words)
    sb = [get_text(link) for link in links]
    sb = [w if len(w) else "x" for w in sb]  # Penalize empty links.
    link_words = [word for lk in sb for word in lk.split() if word.isalnum()]
    num_link_words = len(link_words)
    num_links = len(links)

    proportion = num_link_words * 100 / total_words

    # Function that starts from 65-70% for small values ( 0 - 100 words)
    # and drops to 40-35%  for 700-900 > words (flattens at 35%)
    limit_function = lambda x: 87 - 70 / (1.3 + exp(1 - x / 200))  # noqa E731

    if proportion > limit_function(total_words):
        return True

    if total_words < 50 and num_links > 2 and proportion > 50:
        return True  # Penalize short texts with many links

    return False
    # return True if score > 1.0 else False


def get_node_gravity_score(node):
    gravity_score = node.get("gravityScore")
    return 0.0 if gravity_score is None else float(gravity_score)



"""Common exceptions raised by the newspaper package
"""


class ArticleBinaryDataException(Exception):
    """Exception raised for binary data in urls.
    will be raised if allow_binary_content is False.
    """


class ArticleException(Exception):
    """Generic Article Exception thrown by the article package."""



import argparse
import csv
import io
import logging
from pathlib import Path
import sys
from typing import Any, Dict, Optional, List
import newspaper
from newspaper import settings

logger = logging.getLogger(__name__)


def get_arparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Download and parse news articles.")
    url_group = parser.add_mutually_exclusive_group(required=True)

    url_group.add_argument(
        "--url", "-u", type=str, help="The URL of the article to download and parse."
    )
    url_group.add_argument(
        "--urls-from-file",
        "-uf",
        type=str,
        help="The file containing the URLs of the articles to download and parse.",
    )
    url_group.add_argument(
        "--urls-from-stdin", "-us", action="store_true", help="Read URLs from stdin."
    )
    parser.add_argument(
        "--html-from-file",
        "-hf",
        type=str,
        help=(
            "The HTML file to parse. This will not download the article, it will parse"
            " the HTML file directly."
        ),
    )
    parser.add_argument(
        "--language",
        "-l",
        type=str,
        default="en",
        help="The language of the article to download and parse.",
    )
    parser.add_argument(
        "--output-format",
        "-of",
        choices=["csv", "json", "text"],
        default="json",
        help="The output format of the parsed article.",
    )
    parser.add_argument(
        "--output-file", "-o", type=str, help="The file to write the parsed article to."
    )
    parser.add_argument(
        "--read-more-link",
        type=str,
        help=(
            "A xpath selector for the link to the full article for the case where the"
            ' article is only a summary, and you need to press "read-more" to read the'
            " full text."
        ),
    )
    parser.add_argument(
        "--skip-fetch-images",
        action="store_true",
        help=(
            "Whether to skip fetching images when identifying the top image. This"
            " option speeds up parsing, but can lead to erroneous top image"
            " identification."
        ),
    )
    parser.add_argument(
        "--follow-meta-refresh",
        action="store_true",
        help="Whether to follow meta refresh links when downloading the article.",
    )
    parser.add_argument(
        "--browser-user-agent",
        "-ua",
        type=str,
        help="The user agent string to use when downloading the article.",
    )
    parser.add_argument(
        "--proxy",
        type=str,
        help=(
            "The proxy to use when downloading the article. The format is:"
            " http://<proxy_host>:<proxy_port> e.g.: http://10.10.1.1:8080"
        ),
    )
    parser.add_argument(
        "--request-timeout",
        type=int,
        default=7,
        help="The timeout to use when downloading the article.",
    )
    parser.add_argument(
        "--cookies",
        type=str,
        help=(
            "The cookies to use when downloading the article. The format is:"
            " cookie1=value1; cookie2=value2; ..."
        ),
    )
    parser.add_argument(
        "--skip-ssl-verify",
        action="store_true",
        help="Whether to skip the certificate verification for the article URL ",
    )
    parser.add_argument(
        "--max-nr-keywords",
        type=int,
        default=10,
        help="The maximum number of keywords to extract from the article.",
    )
    parser.add_argument(
        "--skip-nlp", action="store_true", help="Whether to skip the NLP step."
    )

    return parser


def get_kwargs(args: argparse.Namespace) -> dict:
    res: Dict[str, Any] = {}
    if args.html_from_file:
        if Path(args.html_from_file).exists():
            res["input_html"] = Path(args.html_from_file).read_text(encoding="utf-8")
        else:
            logger.error("The file %s does not exist.", args.html_from_file)
            raise FileNotFoundError(f"The file {args.html_from_file} does not exist.")

    if args.language:
        res["language"] = args.language
    if args.read_more_link:
        res["read_more_link"] = args.read_more_link
    if args.skip_fetch_images:
        res["fetch_images"] = False
    if args.follow_meta_refresh:
        res["follow_meta_refresh"] = True
    if args.browser_user_agent:
        res["browser_user_agent"] = args.browser_user_agent
    if args.proxy:
        res["request_kwargs"] = {"proxies": {"http": args.proxy, "https": args.proxy}}
    if args.request_timeout:
        res["request_timeout"] = args.request_timeout
    if args.skip_ssl_verify:
        res["verify"] = False
    if args.cookies:
        cookie_dict = {}
        for cookie in args.cookies.split(";"):
            key, value = cookie.split("=")
            cookie_dict[key.strip()] = value.strip()
        res["cookies"] = cookie_dict

    if args.max_nr_keywords:
        res["max_keywords"] = args.max_nr_keywords

    return res


def run(args: argparse.Namespace):
    if args.html_from_file and (args.urls_from_file or args.urls_from_stdin):
        logger.warning(
            "You specified --html-from-file, but also --urls-from-file or"
            " --urls-from-stdin. The --html-from-file option will be used just for the"
            " first URL."
        )

    if args.urls_from_file:
        with open(args.urls_from_file, "r", encoding="utf-8") as f:
            urls = f.readlines()
    elif args.urls_from_stdin:
        urls = sys.stdin.readlines()
    else:
        urls = [args.url]

    kwargs = get_kwargs(args)

    # Create empty output file. Any existing data will be overwritten.
    if args.output_file:
        logger.info("Writing output to file: %s", args.output_file)
        with open(args.output_file, "w", encoding="utf-8") as f:
            _ = f.write("[") if args.output_format == "json" else f.write("")

    # write one article record (we can have multiple urls)
    def write_output(txt: str):
        if args.output_file:
            with open(args.output_file, "a", encoding="utf-8") as f:
                f.write(txt)
        else:
            print(txt)

    # Create the csv header
    def csv_header() -> str:
        with io.StringIO() as f:
            writer = csv.DictWriter(f, fieldnames=settings.article_json_fields)
            writer.writeheader()
            return f.getvalue()

    # create a csv line as string
    def csv_string(article_dict: dict) -> str:
        with io.StringIO() as f:
            writer = csv.DictWriter(f, fieldnames=settings.article_json_fields)
            writer.writerow(article_dict)
            return f.getvalue()

    for idx, url in enumerate(urls):
        logger.info("Parsing article %d of %d: %s", idx + 1, len(urls), url)
        article = newspaper.article(url=url, **kwargs)

        if not args.skip_nlp:
            article.nlp()
        else:
            logger.info("Skipping NLP step.")

        if "input_html" in kwargs:
            #  we use the html_from_file option just for the first URL
            del kwargs["input_html"]

        output = article.to_json(as_string=args.output_format == "json")
        if args.output_format == "json":
            if idx < len(urls) - 1:
                output += ","
            write_output(output)
        elif args.output_format == "csv":
            if idx == 0:
                write_output(csv_header())
            write_output(csv_string(output))
        elif args.output_format == "text":
            write_output(f"{article.title}\n\n")
            write_output(article.text)
        else:
            raise ValueError(f"Unknown output format: {args.output_format}")

        if args.output_file and args.output_format == "json":
            with open(args.output_file, "a", encoding="utf-8") as f:
                f.write("]")


def main(argv: Optional[List] = None):
    parser = get_arparse()

    args = parser.parse_args(argv)

    try:
        run(args)
    except KeyboardInterrupt:
        sys.exit("\nERROR: Interrupted by user, exiting...\n")


__all__ = ["main"]



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

"""
To change the version of entire package, just edit this one location.
"""

version_info = (0, 9, 1)
__version__ = ".".join(map(str, version_info))



from typing import Optional


languages_tuples = [
    ("aa", "Afar"),
    ("ab", "Abkhazian"),
    ("af", "Afrikaans"),
    ("ak", "Akan"),
    ("sq", "Albanian"),
    ("am", "Amharic"),
    ("ar", "Arabic"),
    ("an", "Aragonese"),
    ("hy", "Armenian"),
    ("as", "Assamese"),
    ("av", "Avaric"),
    ("ae", "Avestan"),
    ("ay", "Aymara"),
    ("az", "Azerbaijani"),
    ("ba", "Bashkir"),
    ("bm", "Bambara"),
    ("eu", "Basque"),
    ("be", "Belarusian"),
    ("bn", "Bengali"),
    ("bh", "Bihari languages"),
    ("bi", "Bislama"),
    ("bo", "Tibetan"),
    ("bs", "Bosnian"),
    ("br", "Breton"),
    ("bg", "Bulgarian"),
    ("my", "Burmese"),
    ("ca", "Catalan; Valencian"),
    ("cs", "Czech"),
    ("ch", "Chamorro"),
    ("ce", "Chechen"),
    ("zh", "Chinese"),
    (
        "cu",
        "Church Slavic; Old Slavonic;",
    ),
    ("cv", "Chuvash"),
    ("kw", "Cornish"),
    ("co", "Corsican"),
    ("cr", "Cree"),
    ("cy", "Welsh"),
    ("cs", "Czech"),
    ("da", "Danish"),
    ("de", "German"),
    ("dv", "Divehi; Dhivehi; Maldivian"),
    ("nl", "Dutch; Flemish"),
    ("dz", "Dzongkha"),
    ("el", "Greek, Modern (1453-)"),
    ("en", "English"),
    ("eo", "Esperanto"),
    ("et", "Estonian"),
    ("eu", "Basque"),
    ("ee", "Ewe"),
    ("fo", "Faroese"),
    ("fa", "Persian"),
    ("fj", "Fijian"),
    ("fi", "Finnish"),
    ("fr", "French"),
    ("fy", "Western Frisian"),
    ("ff", "Fulah"),
    ("Ga", "Georgian"),
    ("de", "German"),
    ("gd", "Gaelic; Scottish Gaelic"),
    ("ga", "Irish"),
    ("gl", "Galician"),
    ("gv", "Manx"),
    ("el", "Greek, Modern (1453-)"),
    ("gn", "Guarani"),
    ("gu", "Gujarati"),
    ("ht", "Haitian; Haitian Creole"),
    ("ha", "Hausa"),
    ("he", "Hebrew"),
    ("hz", "Herero"),
    ("hi", "Hindi"),
    ("ho", "Hiri Motu"),
    ("hr", "Croatian"),
    ("hu", "Hungarian"),
    ("hy", "Armenian"),
    ("ig", "Igbo"),
    ("is", "Icelandic"),
    ("io", "Ido"),
    ("ii", "Sichuan Yi; Nuosu"),
    ("iu", "Inuktitut"),
    ("ie", "Interlingue; Occidental"),
    ("ia", "Interlingua (International Auxiliary Language Association)"),
    ("id", "Indonesian"),
    ("ik", "Inupiaq"),
    ("is", "Icelandic"),
    ("it", "Italian"),
    ("jv", "Javanese"),
    ("ja", "Japanese"),
    ("kl", "Kalaallisut; Greenlandic"),
    ("kn", "Kannada"),
    ("ks", "Kashmiri"),
    ("ka", "Georgian"),
    ("kr", "Kanuri"),
    ("kk", "Kazakh"),
    ("km", "Central Khmer"),
    ("ki", "Kikuyu; Gikuyu"),
    ("rw", "Kinyarwanda"),
    ("ky", "Kirghiz; Kyrgyz"),
    ("kv", "Komi"),
    ("kg", "Kongo"),
    ("ko", "Korean"),
    ("kj", "Kuanyama; Kwanyama"),
    ("ku", "Kurdish"),
    ("lo", "Lao"),
    ("la", "Latin"),
    ("lv", "Latvian"),
    ("li", "Limburgan; Limburger; Limburgish"),
    ("ln", "Lingala"),
    ("lt", "Lithuanian"),
    ("lb", "Luxembourgish; Letzeburgesch"),
    ("lu", "Luba-Katanga"),
    ("lg", "Ganda"),
    ("mk", "Macedonian"),
    ("mh", "Marshallese"),
    ("ml", "Malayalam"),
    ("mi", "Maori"),
    ("mr", "Marathi"),
    ("ms", "Malay"),
    ("Mi", "Micmac"),
    ("mk", "Macedonian"),
    ("mg", "Malagasy"),
    ("mt", "Maltese"),
    ("mn", "Mongolian"),
    ("mi", "Maori"),
    ("ms", "Malay"),
    ("my", "Burmese"),
    ("na", "Nauru"),
    ("nv", "Navajo; Navaho"),
    ("nr", "Ndebele, South; South Ndebele"),
    ("nd", "Ndebele, North; North Ndebele"),
    ("ng", "Ndonga"),
    ("ne", "Nepali"),
    ("nl", "Dutch; Flemish"),
    ("nn", "Norwegian Nynorsk; Nynorsk, Norwegian"),
    ("nb", "Bokmål, Norwegian; Norwegian Bokmål"),
    ("no", "Norwegian"),
    ("oc", "Occitan (post 1500)"),
    ("oj", "Ojibwa"),
    ("or", "Oriya"),
    ("om", "Oromo"),
    ("os", "Ossetian; Ossetic"),
    ("pa", "Panjabi; Punjabi"),
    ("fa", "Persian"),
    ("pi", "Pali"),
    ("pl", "Polish"),
    ("pt", "Portuguese"),
    ("ps", "Pushto; Pashto"),
    ("qu", "Quechua"),
    ("rm", "Romansh"),
    ("ro", "Romanian; Moldavian; Moldovan"),
    ("ro", "Romanian; Moldavian; Moldovan"),
    ("rn", "Rundi"),
    ("ru", "Russian"),
    ("sg", "Sango"),
    ("sa", "Sanskrit"),
    ("si", "Sinhala; Sinhalese"),
    ("sk", "Slovak"),
    ("sk", "Slovak"),
    ("sl", "Slovenian"),
    ("se", "Northern Sami"),
    ("sm", "Samoan"),
    ("sn", "Shona"),
    ("sd", "Sindhi"),
    ("so", "Somali"),
    ("st", "Sotho, Southern"),
    ("es", "Spanish; Castilian"),
    ("sq", "Albanian"),
    ("sc", "Sardinian"),
    ("sr", "Serbian"),
    ("ss", "Swati"),
    ("su", "Sundanese"),
    ("sw", "Swahili"),
    ("sv", "Swedish"),
    ("ty", "Tahitian"),
    ("ta", "Tamil"),
    ("tt", "Tatar"),
    ("te", "Telugu"),
    ("tg", "Tajik"),
    ("tl", "Tagalog"),
    ("th", "Thai"),
    ("bo", "Tibetan"),
    ("ti", "Tigrinya"),
    ("to", "Tonga (Tonga Islands)"),
    ("tn", "Tswana"),
    ("ts", "Tsonga"),
    ("tk", "Turkmen"),
    ("tr", "Turkish"),
    ("tw", "Twi"),
    ("ug", "Uighur; Uyghur"),
    ("uk", "Ukrainian"),
    ("ur", "Urdu"),
    ("uz", "Uzbek"),
    ("ve", "Venda"),
    ("vi", "Vietnamese"),
    ("vo", "Volapük"),
    ("cy", "Welsh"),
    ("wa", "Walloon"),
    ("wo", "Wolof"),
    ("xh", "Xhosa"),
    ("yi", "Yiddish"),
    ("yo", "Yoruba"),
    ("za", "Zhuang; Chuang"),
    ("zh", "Chinese"),
    ("zu", "Zulu"),
]

languages_dict = {code: language for code, language in languages_tuples}


def get_language_from_iso639_1(iso639_1: str) -> Optional[str]:
    """Returns the long language name from the iso639_1 code

    Args:
        iso639_1 (str): the two character iso639_1 code

    Returns:
        str: Language name (in english)
    """
    return languages_dict.get(iso639_1)



# -*- coding: utf-8 -*-
# Much of the code here was forked from https://github.com/codelucas/newspaper
# Copyright (c) Lucas Ou-Yang (codelucas)

"""
Global package-wide settings and constants live here.
"""
import logging
from pathlib import Path
import tempfile

from .version import __version__

log = logging.getLogger(__name__)

NUM_THREADS_PER_SOURCE_WARN_LIMIT = 5

# Expected value for sentence length
MEAN_SENTENCE_LEN = 20.0
SUMMARIZE_KEYWORD_COUNT = 10

PARENT_DIRECTORY = Path(__file__).resolve().parent
POPULAR_URLS = PARENT_DIRECTORY / "resources/misc/popular_sources.txt"
USERAGENTS = PARENT_DIRECTORY / "resources/misc/useragents.txt"
STOPWORDS_DIR = PARENT_DIRECTORY / "resources/text"

DATA_DIRECTORY = ".newspaper_scraper"

TOP_DIRECTORY = Path(tempfile.gettempdir()).resolve() / DATA_DIRECTORY

# Fields used to construct the article json string.

article_json_fields = [
    "url",
    "read_more_link",
    "language",
    "title",
    "top_image",
    "meta_img",
    "images",
    "movies",
    "keywords",
    "meta_keywords",
    "tags",
    "authors",
    "publish_date",
    "summary",
    "meta_description",
    "meta_lang",
    "meta_favicon",
    "meta_site_name",
    "canonical_link",
    "text",
    "text_cleaned",
]

# Tags we allow to be left in the cleaned article body
CLEAN_ARTICLE_TAGS = [
    "a",
    "span",
    "p",
    "br",
    "strong",
    "b",
    "em",
    "i",
    "tt",
    "code",
    "pre",
    "blockquote",
    "img",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "ul",
    "ol",
    "li",
    "dl",
    "dt",
    "dd",
]
# Tags that get a newline appended to them during conversion to text
BLOCK_LEVEL_TAGS = [
    "address",
    "article",
    "aside",
    "blockquote",
    "canvas",
    "dd",
    "div",
    "dl",
    "dt",
    "fieldset",
    "figcaption",
    "figure",
    "footer",
    "form",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "header",
    "hr",
    "li",
    "main",
    "nav",
    "noscript",
    "ol",
    "p",
    "pre",
    "section",
    "table",
    "tfoot",
    "ul",
    "video",
]

# Tags that we remove from cleaned article HTML
ARTICLE_STRIP_TAGS = ["a", "b", "strong", "i", "br", "sup"]

# Known Advertisement Attributes and IDs
ADVERTISEMENT_ATTR_VALUES = r"inline-ads|promotion|sponsor|banner"

# Error log
LOGFILE = TOP_DIRECTORY / f"newspaper_errors_{__version__}.log"
MONITOR_LOGFILE = TOP_DIRECTORY / "newspaper_monitors_{__version__}.log"

# Memo directory (same for all concur crawlers)
MEMO_FILE = "memoized"
MEMO_DIR = TOP_DIRECTORY / MEMO_FILE

# category cache
CACHE_DIRECTORY = TOP_DIRECTORY / "category_cache"

TRENDING_URL = "http://www.google.com/trends/hottrends/atom/feed?pn=p1"

for path in (TOP_DIRECTORY, MEMO_DIR, CACHE_DIRECTORY):
    path.mkdir(parents=True, exist_ok=True)



import logging
from datetime import datetime
from typing import Dict, Any, List, Optional
import lxml
from newspaper import urls
import newspaper.parsers as parsers
from newspaper.configuration import Configuration
from newspaper.extractors.articlebody_extractor import ArticleBodyExtractor
from newspaper.extractors.authors_extractor import AuthorsExtractor
from newspaper.extractors.categories_extractor import CategoryExtractor
from newspaper.extractors.image_extractor import ImageExtractor
from newspaper.extractors.metadata_extractor import MetadataExtractor
from newspaper.extractors.pubdate_extractor import PubdateExtractor
from newspaper.extractors.title_extractor import TitleExtractor
from newspaper.extractors.videos_extractor import VideoExtractor
from newspaper.utils.classes import Video

log = logging.getLogger(__name__)


class ContentExtractor:
    def __init__(self, config: Configuration):
        self.config = config
        self.language = config.language
        self.stopwords_class = config.stopwords_class
        self.title_extractor = TitleExtractor(config)
        self.author_extractor = AuthorsExtractor(config)
        self.pubdate_extractor = PubdateExtractor(config)
        self.atricle_body_extractor = ArticleBodyExtractor(config)
        self.metadata_extractor = MetadataExtractor(config)
        self.categories_extractor = CategoryExtractor(config)
        self.image_extractor = ImageExtractor(config)
        self.video_extractor = VideoExtractor(config)

    def update_language(self, meta_lang: str):
        """Required to be called before the extraction process in some
        cases because the stopwords_class has to set in case the lang
        is not latin based
        """
        if meta_lang:
            self.language = meta_lang
            self.stopwords_class = self.config.get_stopwords_class(meta_lang)
            self.atricle_body_extractor.stopwords_class = self.stopwords_class
            self.atricle_body_extractor.language = self.language

    def get_authors(self, doc: lxml.html.Element) -> List[str]:
        """Fetch the authors of the article, return as a list
        Only works for english articles
        """
        return self.author_extractor.parse(doc)

    def get_publishing_date(
        self, url: str, doc: lxml.html.Element
    ) -> Optional[datetime]:
        """Return the article publishing date as datetime object. If no valid
        date could be found, return None. The parser tries to determine the
        date from the following sources (in this order): article url
        (e.g /2019/12/31/article.html), yoast schema and json ld schema, meta
        data from metatags.

        Args:
            url (str): The article url we parse
            doc (lxml.html.Element): The DOM for the whole article page

        Returns:
            Optional[datetime]: a datetime object or None
        """
        return self.pubdate_extractor.parse(url, doc)

    def get_title(self, doc):
        """Fetch the article title and analyze it

        Assumptions:
        - title tag is the most reliable (inherited from Goose)
        - h1, if properly detected, is the best (visible to users)
        - og:title and h1 can help improve the title extraction
        - python == is too strict, often we need to compare filtered
          versions, i.e. lowercase and ignoring special chars

        Explicit rules:
        1. title == h1, no need to split
        2. h1 similar to og:title, use h1
        3. title contains h1, title contains og:title, len(h1) > len(og:title), use h1
        4. title starts with og:title, use og:title
        5. use title, after splitting
        """
        return self.title_extractor.parse(doc)

    def get_feed_urls(self, source_url, categories):
        """Takes a source url and a list of category objects and returns
        a list of feed urls
        """
        total_feed_urls = []
        for category in categories:
            attribs = {"type": "application/rss+xml"}
            feed_elements = parsers.get_tags(category.doc, attribs=attribs)
            feed_urls = [e.get("href") for e in feed_elements if e.get("href")]
            total_feed_urls.extend(feed_urls)

        total_feed_urls = total_feed_urls[:50]
        total_feed_urls = [urls.prepare_url(f, source_url) for f in total_feed_urls]
        total_feed_urls = list(set(total_feed_urls))
        return total_feed_urls

    def get_metadata(self, article_url: str, doc: lxml.html.Element) -> Dict[str, Any]:
        """Parse the article's HTML for any known metadata attributes"""
        return self.metadata_extractor.parse(article_url, doc)

    def parse_images(
        self, article_url: str, doc: lxml.html.Element, top_node: lxml.html.Element
    ):
        """Parse images in an article"""
        self.image_extractor.parse(doc, top_node, article_url)

    def get_category_urls(self, source_url, doc):
        """Inputs source lxml root and source url, extracts domain and
        finds all of the top level urls, we are assuming that these are
        the category urls.
        cnn.com --> [cnn.com/latest, world.cnn.com, cnn.com/asia]
        """
        return self.categories_extractor.parse(source_url, doc)

    @property
    def top_node(self) -> lxml.html.Element:
        """Returns the top node of the article.
        calculate_best_node() must be called first

        Returns:
            lxml.html.Element: The top node containing the article text
        """
        return self.atricle_body_extractor.top_node

    @property
    def top_node_complemented(self) -> lxml.html.Element:
        """The cleaned version of the top node, without any divs, linkstuffing, etc

        Returns:
            lxml.html.Element: deepcopy version of the top node, cleaned
        """
        return self.atricle_body_extractor.top_node_complemented

    def calculate_best_node(
        self, doc: lxml.html.Element
    ) -> Optional[lxml.html.Element]:
        """Extracts the most probable top node for the article text
        based on a variety of heuristics

        Args:
            doc (lxml.html.Element): Root node of the document.
              The search starts from here.
              usually it's the html tag of the web page

        Returns:
            lxml.html.Element: the article top element
            (most probable container of the article text), or None
        """
        self.atricle_body_extractor.parse(doc)

        return self.atricle_body_extractor.top_node

    def get_videos(
        self, doc: lxml.html.Element, top_node: lxml.html.Element
    ) -> List[Video]:
        """Gets video links from article

        Args:
            doc (lxml.html.Element): Top document node. Needed for json-ld
            top_node (lxml.html.Element): Article top node.

        Returns:
            List[str]: list of video urls
        """
        return self.video_extractor.parse(doc, top_node)
